define([
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/_base/array",
    "thornberry/app/Component/_Component",
    "dojo/store/Memory",
    "dojo/dom-construct",
    "dojo/dom-style",
    "thornberry/report/formatter",
    "dijit/_WidgetBase",
    "gridx/Grid",
    "gridx/core/model/cache/Sync",
    "gridx/modules/select/Row",
    "gridx/modules/extendedSelect/Row",
    "gridx/modules/IndirectSelectColumn",
    "gridx/modules/Filter",
    "gridx/modules/SingleSort",
    "gridx/core/model/extensions/FormatSort",
    "gridx/modules/VirtualVScroller",
    "gridx/modules/ColumnResizer",
    "gridx/modules/HiddenColumns",
    "gridx/modules/move/Column",
    "gridx/modules/CellWidget",
    "gridx/modules/Dod",
    "gridx/modules/Bar",
    "gridx/core/model/extensions/FormatSort"
],function(declare,lang,arrayUtils,_Component,Memory,domConstruct,domStyle,formatter,_WidgetBase,Grid,Cache,SelectRow,ExtendedSelectRow,IndirectSelectColumn,Filter,SingleSort,FormatSort,VirtualVScroller,ColumnResizer,HiddenColumns,MoveColumn,CellWidget,Dod,Bar) {
    var gridRowClickHandler = function(event) {
        // Summary:
        //      When connected with the grid's onRowClick event, this causes clicking the checkbox to add the row to those already selected, and clicking anywhere else
        //      to result in only the clicked row being selected
        if (event.columnIndex !== 0 && this.select.row.enabled) {
            var selected = this.select.row.getSelected();
            if (selected.length === 1 && selected[0] === event.rowId) {
                this.select.row.deselectById(event.rowId);
            } else {
                if (!event.ctrlKey) {
                    this.select.row.clear();
                }
                this.select.row.selectById(event.rowId);
            }
        }
    };
    
    return declare([_Component],{
        columns: null, // Must be passed in, and cannot be changed - [{ id:, name: }...]
        selectableRows: false,

        data: null,
        selectedColumns: null,

        filter: null,

        virtualScroller: true,
        autoHeight: false,

        rowDetails: false,
        detailProvider: null, // function(grid, rowId, node, rendered) called when a row is expanded

        onChangeSelection: function(ids) {},

        dataStore: null,
        style: "height:100%",

        sortInitialOrder: null,

        postMixInProperties: function() {
            var t = this;
            t.inherited(arguments);

            t.dataStore = new Memory();
        },

        buildRendering: function() {
            // Summary:
            //      This method is responsible for setting up the content pane's content 
            var t = this;
            t.inherited(arguments);

            var modules = [
                Bar,
                SingleSort,
                Filter,
                ColumnResizer,
                HiddenColumns,
                MoveColumn,
                CellWidget
            ];
            if (t.virtualScroller) {
                modules.push(VirtualVScroller);
            }
            if (t.selectableRows) {
                modules.push(SelectRow);
                modules.push(IndirectSelectColumn);
                modules.push(ExtendedSelectRow);
            }
            if (t.rowDetails && t.detailProvider) {
                modules.push({
                    moduleClass: Dod,
                    defaultShow: false,
                    useAnimation: false,
                    showExpando: false,
                    detailProvider: t.detailProvider
                });
            }

            t.topBar = declare(_WidgetBase,{
                text: "",
                _setTextAttr: { node: "domNode", type: "innerHTML" },
                buildRendering: function() {
                    this.inherited(arguments);
                    this.domNode = domConstruct.toDom("<div class='emphasize' style='padding:3px'></div>");
                }
            })();

            t.grid = new Grid({
                gridView: t,
                cacheClass: Cache,
                store: t.dataStore,
                structure: t.generateGridStructure(t.columns),
                modules: modules,
                modelExtensions: [
                    FormatSort
                ],
                sortInitialOrder: t.sortInitialOrder,
                autoHeight: t.autoHeight,
                barTop: [
                    t.topBar
                ],
                modelExtensions: [
                    "gridx/core/model/extensions/FormatSort"
                ]
            });
            
            t.domNode = t.grid.domNode;
            t.grid.startup();
            t.grid.resize();
            
            //t.grid.connect(t.grid.columnResizer,"onResize",lang.hitch(t,"saveColumnWidth"));
            if (t.grid.select) {
                t.grid.connect(t.grid,"onRowClick",gridRowClickHandler);
                t.grid.connect(t.grid.select.row,"onSelectionChange",function() {t.onChangeSelection(t.getSelected());});
            }

           domStyle.set(t.grid.bar.topNode,"display","none");
        },

        _setSelectedColumnsAttr: function(value) {
            var t = this;
            t.setColumns(value);
        },

        _setFilterAttr: function(value) {
            var t = this;
            t.filter = value;
            t.refresh();
        },

        _setDataAttr: function(data) {
            var t = this;
            t.dataStore.setData(data);
            t.refresh();
        },

        setColumns: function(columns) {
            // Summary:
            //      Takes an array of the columns that the user has selected to show, rearranges the grid columns to match, and hides any not included
            var t = this;
            var grid = t.grid;
            grid.hiddenColumns.add.apply(grid.hiddenColumns,grid.columns());
            grid.hiddenColumns.remove.apply(grid.hiddenColumns,columns);

            // Clear sorting if the sort column is no longer displayed
            if (!arrayUtils.some(columns,function(column) {
                return column == t.grid.sort._sortId;
            })) {
                t.grid.sort._sortId = "";
            }

            var columnPosition = 1;
            arrayUtils.forEach(columns,function(id) {
                var column = grid.column(id);
                if (column) {
                    grid.move.column.move(column.index(),columnPosition);
                    columnPosition++;
                }
            });
        },

        saveColumnWidth: function(column,width) {
            // Summary:
            //      Saves a column width in the user's settings on the server
            this.controller.userSettings.set("columnWidth",column,width,true);
        },

        refresh: function() {
            // Summary:
            //      After the store's data has been changed, this synchronizes the grid
            var t = this;
            if (t.grid.select && t.grid.select.row) {
                t.grid.select.row.clear();
            }
            t.grid.vScroller.scrollToRow(0,1);
            t.grid.model.clearCache();
            if (t.filter) {
                t.grid.model.filter(t.filter);
            }
            t.grid.body.refresh(); 
            t.grid.model.onSizeChange();
        },

        generateGridStructure: function(columns) {
            // Summary:
            //      Takes a thornberry/report/Report and generates the column structure from it
            var t = this;
            var structure = [];
            arrayUtils.forEach(columns,function(columnDefinition) {
                // Set basic column data
                var displayName = escapeText(columnDefinition.name);

                var gridColumn = {
                    id: columnDefinition.id,
                    name: displayName,
                    field: columnDefinition.id,
                    formatter: columnDefinition.formatter,
                    decorator: columnDefinition.decorator,
                    widgetsInCell: columnDefinition.widgetsInCell,
                    setCellValue: columnDefinition.setCellValue,
                    sortFormatted: columnDefinition.sortFormatted
                };

                // Get the column formatter
                //gridColumn.formatter = formatter.getForColumn(gridColumn,columnDefinition);

                // Set the column's default width
                var width = ""; //t.controller.userSettings.get("columnWidth",columnDefinition.name);
                width = (width !== "" ? width : columnDefinition.defaultWidth || "");
                if (width !== "") {
                    gridColumn.width = width + "px";
                }

                structure.push(gridColumn);
            });

            return structure;
        },

        getSelected: function() {
            var t = this;
            var selectedIds = [];
            var rows = t.grid.rows();
            for (var i = 0; i < rows.length; i++) {
                if (t.grid.select.row.isSelected(rows[i].id)) {
                    selectedIds.push(rows[i].id);
                }
            }
            return selectedIds;
        },

        getAll: function() {
            return arrayUtils.map(this.grid.rows(),function(row) {
                return row.id;
            });
        },

        resize: function() {
            this.grid.resize();
        },

        setTopBarText: function(text) {
            var t = this;
            if (text) {
                t.topBar.set("text",text);
                domStyle.set(t.grid.bar.topNode,"display","");
            } else {
                domStyle.set(t.grid.bar.topNode,"display","none");
            }
        }
    });
});