define([
  "dojo/_base/declare",
  "dojo/_base/lang",
  "dojo/_base/array",
  "dojo/on",
  "dojox/calendar/CalendarBase",
  "thornberry/calendar/VerticalRenderer",
  "thornberry/calendar/VerticalSideRenderer",
  "thornberry/calendar/ColumnView",
  "dojox/calendar/ColumnViewSecondarySheet",
  "dojox/calendar/DecorationRenderer",
  "thornberry/calendar/MatrixView",
  "thornberry/calendar/AgendaView",
  "thornberry/calendar/SideBySideView",
  "thornberry/calendar/SideBySideSecondarySheet",
  "thornberry/calendar/HorizontalRenderer",
  "thornberry/calendar/LabelRenderer",
  "dojox/calendar/ExpandRenderer",
  "thornberry/calendar/AgendaRenderer",
  "dojox/calendar/Mouse",
  "dojox/calendar/Touch",
  "dojox/calendar/Keyboard",
  "dojo/dom-style",
  "dojo/text!./template.html",
  "dojo/_base/array",
  "dijit/form/Button",
  "dijit/Toolbar",
  "dijit/ToolbarSeparator",
], function (
  declare,
  lang,
  arr,
  on,
  CalendarBase,
  VerticalRenderer,
  VerticalSideRenderer,
  ColumnView,
  ColumnViewSecondarySheet,
  DecorationRenderer,
  MatrixView,
  AgendaView,
  SideBySideView,
  SideBySideSecondarySheet,
  HorizontalRenderer,
  LabelRenderer,
  ExpandRenderer,
  AgendaRenderer,
  Mouse,
  CalendarTouch,
  Keyboard,
  domStyle,
  template,
  arrayUtils
) {
  return declare([CalendarBase], {
    // summary:
    //		A Calendar subclass that embeds a month column view.

    templateString: template,

    verticalRenderer: VerticalRenderer,

    animateRange: false,

    summaryView: false,

    sideBySideViewClicked: false,

    calendarView: null, // Containing object

    dateInterval: "month",

    postMixInProperties: function () {
      var t = this;
      t.inherited(arguments);

      t.invalidatingProperties.push("summaryView");
      t.invalidatingProperties.push("sideBySideViewClicked");
      t.invalidatingProperties.push("subColumns");
    },

    formatItemTimeFunc: function (d, rd, view, item) {
      return rd.dateLocaleModule
        .format(d, {
          selector: "time",
          timePattern: d.getMinutes() == 0 ? "ha" : "h:mma",
        })
        .toLowerCase();
    },

    isItemEditable: function (item, rendererKind) {
      return item.event
        ? item.event.editable && item.event.dragAndDropEnabled
        : item._item.event.editable && item._item.event.dragAndDropEnabled;
    },

    _createDefaultViews: function () {
      this.inherited(arguments);
	  
	  var secondaryViewOptions = [ColumnViewSecondarySheet, Keyboard, Mouse];
	  if (IS_MOBILE) {
		secondaryViewOptions.push(CalendarTouch);
	  }
		
      var secondarySheetClass = declare(secondaryViewOptions,
        {
          roundToDay: true,

          _sortItemsFunction: function (a, b) {
            var t = this;
            return t.inherited(arguments);

            /*if(this.roundToDay){
						a = this._roundItemToDay(a);
						b = this._roundItemToDay(b);
					}
					var res = this.dateModule.compare(a.startTime, b.startTime);
					if(res == 0){
						res = -1 * this.dateModule.compare(a.endTime, b.endTime);
					}
					return res;*/
          },
        }
      );
	  
	  var colViewOptions = [ColumnView, Keyboard, Mouse];
	  if (IS_MOBILE) {
		colViewOptions.push(CalendarTouch);
	  }
		
      var colView = declare(colViewOptions)(
        lang.mixin(
          {
            secondarySheetClass: secondarySheetClass,
            verticalRenderer: VerticalRenderer,
            horizontalRenderer: HorizontalRenderer,
            expandRenderer: ExpandRenderer,
            horizontalDecorationRenderer: DecorationRenderer,
            verticalDecorationRenderer: DecorationRenderer,
            percentOverlap: 0,
            minHours: 0,
            maxHours: 24,
            startTimeOfDay: { hours: new Date().getHours() },
            _formatColumnHeaderLabel: function (d) {
              return this.renderData.dateLocaleModule.format(d, {
                selector: "date",
                datePattern: "EEE M/d",
              });
            },
          },
          this.columnViewProps
        )
      );

	  var matrixViewOptions = [MatrixView, Keyboard, Mouse];
	  if (IS_MOBILE) {
		matrixViewOptions.push(CalendarTouch);
	  }

      var matrixView = declare(matrixViewOptions)(
        lang.mixin(
          {
            horizontalRenderer: HorizontalRenderer,
            horizontalDecorationRenderer: DecorationRenderer,
            labelRenderer: HorizontalRenderer,
            expandRenderer: ExpandRenderer,
            labelRendererHeight: 17,
            _formatColumnHeaderLabel: function (d) {
              var value = "";
              try {
                value = this.renderData.dateLocaleModule.getNames(
                  "days",
                  this.columnHeaderLabelLength
                    ? this.columnHeaderLabelLength
                    : "wide",
                  "standAlone"
                )[d.getDay()];
              } catch (e) {}
              return value;
            },
          },
          this.matrixViewProps
        )
      );

	  var agendaViewOptions = [AgendaView, Keyboard, Mouse];
	  if (IS_MOBILE) {
		agendaViewOptions.push(CalendarTouch);
	  }
		
      var agendaView = declare(agendaViewOptions)({
        agendaRenderer: AgendaRenderer,
      });

      var t = this;

	  var sideBySideSecondarySheetViewOptions = [SideBySideSecondarySheet, Keyboard, Mouse];
	  if (IS_MOBILE) {
		sideBySideSecondarySheetViewOptions.push(CalendarTouch);
	  }

      var sideBySideSecondarySheetClass = declare(sideBySideSecondarySheetViewOptions,
        {
          roundToDay: true,

          horizontalRenderer: HorizontalRenderer,

          _sortItemsFunction: function (a, b) {
            var t = this;
            return t.inherited(arguments);

            /*if(this.roundToDay){
						a = this._roundItemToDay(a);
						b = this._roundItemToDay(b);
					}
					var res = this.dateModule.compare(a.startTime, b.startTime);
					if(res == 0){
						res = -1 * this.dateModule.compare(a.endTime, b.endTime);
					}
					return res;*/
          },
        }
      );

	  var sideBySideViewOptions = [SideBySideView, Keyboard, Mouse];
	  if (IS_MOBILE) {
		sideBySideViewOptions.push(CalendarTouch);
	  }

      var sideBySideView = declare(sideBySideViewOptions)({
        secondarySheetClass: sideBySideSecondarySheetClass,
        verticalRenderer: VerticalSideRenderer,
        expandRenderer: ExpandRenderer,
        horizontalDecorationRenderer: DecorationRenderer,
        verticalDecorationRenderer: DecorationRenderer,
        percentOverlap: 0,
        calendarStore: this.calendarStore,
        minHours: 0,
        maxHours: 24,
        startTimeOfDay: { hours: new Date().getHours() },
        _formatColumnHeaderLabel: function (d) {
          return this.renderData.dateLocaleModule.format(d, {
            selector: "date",
            datePattern: "EEE M/d",
          });
        },

        subColumns: arrayUtils.map(
          this.calendarStore.query({ type: "Employee" }),
          function (calendar) {
            return calendar.id;
          }
        ),
        subColumnLabelFunc: function (v) {
          var calendar = t.calendarStore.get(v);
          if (calendar) {
            return (
              calendar.name.toString().replace(/,/g, ", ") +
              " " +
              calendar.identifier +
              " <a onclick='showDetails(\"" +
              calendar.type.toLowerCase() +
              '","' +
              calendar.calendarId +
              "\")'><i class='fas fa-info-circle'></i></a>"
            );
          }
          return null;
        },

        itemToRendererKindFunc: function (item) {
          return item.allDay || item._item.calendar.type !== "Employee"
            ? null
            : "vertical"; // String
        },
      });

      this.columnView = colView;
      this.matrixView = matrixView;
      this.agendaView = agendaView;
      this.sideBySideView = sideBySideView;

      this.matrixView.on(
        "rowHeaderClick",
        lang.hitch(this, this.matrixViewRowHeaderClick)
      );
      this.columnView.on(
        "columnHeaderClick",
        lang.hitch(this, this.columnViewColumnHeaderClick)
      );
      this.agendaView.on(
        "rowHeaderClick",
        lang.hitch(this, this.agendaViewRowHeaderClick)
      );

      return [
        this.columnView,
        this.matrixView,
        this.agendaView,
        this.sideBySideView,
      ];
    },

    _setDateIntervalAttr: function (value) {
      var t = this;
      t.inherited(arguments);
      t.set("summaryView", false);
      t.set("sideBySideViewClicked", false);
    },

    _setSummaryViewAttr: function (value) {
      var t = this;
      if (t.summaryView !== value) {
        t._timeInterval = null;
      }
      t._set("summaryView", value);
      t._timeRangeInvalidated = true;
    },

    _setSideBySideViewClickedAttr: function (value) {
      var t = this;
      if (t.sideBySideViewClicked !== value) {
        t._timeInterval = null;
      }
      t._set("sideBySideViewClicked", value);
      t._timeRangeInvalidated = true;
    },

    _setSubColumnsAttr: function (value) {
      var t = this;
      t._set("subColumns", value);
      t.sideBySideView.set("subColumns", value);
    },

    _computeCurrentView: function (startDate, endDate, duration) {
      var t = this;
      var view = null;
      if (t.summaryView) {
        view = this.agendaView;
      } else if (t.sideBySideViewClicked) {
        view = this.sideBySideView;
      } else {
        view = t.inherited(arguments);
      }
      return view;
    },

    _configureView: function (view, index, timeInterval, duration) {
      if (view.viewKind == "agenda") {
        view.set("startDate", this.newDate(timeInterval[0]));
        view.set("rowCount", duration);
      } else {
        this.inherited(arguments);
      }
    },

    _navigate: function (dir) {
      var t = this;

      if (t.calendarView.dropDownButton.label === "Cert") {
        var calendar = t.calendarView.patientCalendar;
        var currentDate = this.get("date") || this.get("startDate");
        currentCert = calendar.getCertByDate(currentDate);
        certIndex = currentCert.counter - 1 + dir;
        if (certIndex < 0) {
          certIndex = 0;
        }
        if (certIndex > calendar.other.certs.length - 1) {
          certIndex = calendar.other.certs.length - 1;
        }
        var cert = calendar.other.certs[certIndex];
        t.set("date", null);
        t.set("startDate", new Date(cert.start));
        t.set("endDate", new Date(cert.end));
      } else {
        t.inherited(arguments);
      }
    },

    computeTimeInterval: function () {
      var t = this;
      if (
        t.calendarView.dropDownButton.label !== "Cert" &&
        !t.get("date") &&
        t.get("startDate")
      ) {
        t.set("date", t.get("startDate"));
      }
      return t.inherited(arguments);
    },

    _computeTimeIntervalImpl: function (d) {
      var t = this;
      var interval = null;
      if (t.calendarView.dropDownButton.label === "Cert") {
        var calendar = t.calendarView.patientCalendar;
        var cert = calendar.getCertByDate(t.floorToDay(d));
        interval = [new Date(cert.start), new Date(cert.end)];
        t.date = null;
        t.startDate = new Date(cert.start);
        t.endDate = new Date(cert.end);
      } else {
        return t.inherited(arguments);
      }
      return interval;
    },

    configureButtons: function () {
      this.inherited(arguments);
      this.own(
        on(
          this.twoWeekButton,
          "click",
          lang.hitch(this, function () {
            this.set("dateIntervalSteps", 2);
            this.set("dateInterval", "week");
          })
        ),
        on(
          this.certButton,
          "click",
          lang.hitch(this, function () {
            this.set("dateIntervalSteps", 90);
            this.set("dateInterval", "day");
          })
        ),
        on(
          this.agendaButton,
          "click",
          lang.hitch(this, function () {
            this.set("dateIntervalSteps", 1);
            this.set("dateInterval", "week");
            this.set("summaryView", true);
          })
        ),
        on(
          this.sideBySideButton,
          "click",
          lang.hitch(this, function () {
            this.set("dateIntervalSteps", 1);
            this.set("dateInterval", "day");
            this.set("sideBySideViewClicked", true);
          })
        )
      );
    },

    goToday: function () {
      this.set("date", this.floorToDay(new this.dateClassObj(), true));
    },

    matrixViewRowHeaderClick: function (e) {
      this.calendarView.dropDownButton.set("label", "Week");
      this.set("dateInterval", "week");
      this.set("dateIntervalSteps", 1);
      this.set("date", e.date);
    },

    columnViewColumnHeaderClick: function (e) {
      // summary:
      //		Function called when the cell of a column header of the column view is clicked.
      //		Show the time range defined by the clicked date.
      // e: Object
      //		The column header click event.
      // tags:
      //		protected

      var cal = this.dateModule;
      if (
        cal.compare(e.date, this._timeInterval[0]) == 0 &&
        this.dateInterval == "day" &&
        this.dateIntervalSteps == 1
      ) {
        this.calendarView.dropDownButton.set("label", "Week");
        this.set("dateInterval", "week");
      } else {
        this.calendarView.dropDownButton.set("label", "Day");
        this.set("date", e.date);
        this.set("dateInterval", "day");
        this.set("dateIntervalSteps", 1);
      }
    },

    agendaViewRowHeaderClick: function (e) {
      this.calendarView.dropDownButton.set("label", "Day");
      this.set("dateInterval", "day");
      this.set("dateIntervalSteps", 1);
      this.set("date", e.date);
    },

    dayButtonClick: function (e) {
      // summary:
      //		The action triggerred when the day button is clicked.
      //		By default, sets the dateInterval property to "day" and
      //		the "dateIntervalSteps" to 1.

      if (this.get("date") == null) {
        if (this.selectedItem) {
          this.set("date", new Date(this.selectedItem._startTime));
        } else if (this.startDate) {
          this.set("date", this.startDate);
        } else {
          this.set("date", this.floorToDay(new this.dateClassObj(), true));
        }
      }
      this.set("dateInterval", "day");
      this.set("dateIntervalSteps", 1);
    },
  });
});
