define([
    "dojo/_base/declare",
    "dojo/_base/lang",
    "dojo/Stateful",
    "dojo/date",
    "dojo/date/locale",
    "dojo/aspect",
    "dojo/_base/array",
    "dojo/Deferred",
    "thornberry/serverRequest",
    "thornberry/editor/load",
    "dojo/when",
    "dojo/dom-construct",
    "thornberry/app",
    "dijit/registry",
    "dojo/when",
    "dojo/dom-style",
    "dojo/store/Memory",
    "dojo/topic"
], function (declare,lang,Stateful,dateUtils,locale,aspect,arrayUtils,Deferred,serverRequest,load,when,domConstruct,app,registry,when,domStyle,Memory,topic) {
    var lastId = 1;
    
    var Event =  declare([Stateful], {
        id: "",
        eventId: "",
        className: "",
        date: "",
        time: "",
        duration: 0,
        division: "",
        type: "",
        description: "", // remove and replace in other
        status: "",
        editable: "",
        calendars: null,
        travelTime: 0,

        _startTime: null,
        _endTime: null,
        _allDay: false,
        _statusClass: "",
        _description: "",
        
        endDate: "",
        details: null,

        calendarEventArray: null,

        calendarStore: {},

        postscript: function() {
            var t = this;
            t.inherited(arguments);

            if (t.id === "") {
                t.id = lastId++;
            }

            if (t._endTime !== "") {
                t.endDate = locale.format(new Date(t._endTime),{selector:"date",datePattern:"yyyy-MM-dd"});
            }

            t._description = t.type;
            if (t.division !== "") {
                t._description = t.division + " " + t._description;
            }

            t._statusClass = t.getStatusClass();
        },

        getCalendarEvents: function(calendarIndex,calendarStore) {
            var t = this;
            t.calendarEventArray = [];
            arrayUtils.forEach(t.calendarIds,function(calendarId) {
                if (calendarIndex[calendarId]) {
                    var calendar = calendarStore.get(calendarId);
                    var displayedCalendar = t.getCalendarForDisplay(calendar);
                    t.calendarEventArray.push({
                        id: calendarId+"||"+t.id,
                        _startTime: t._startTime,
                        _endTime: t._endTime,
                        _allDay: t._allDay,
                        _description: t._description,
                        primaryDetails: (typeof displayedCalendar === "object" && displayedCalendar !== null ? displayedCalendar.name : displayedCalendar),
                        secondaryDetails: t.description,
                        event: t,
                        calendar: calendar, // Get the calendar set up in the store with color data
                        _calendarId: calendarId,
                        _calendarIdOrig: calendarId
                    });
                }
            });
            return t.calendarEventArray;
        },

        propagateChanges: function(calendarEvent,calendar,calendarStore) {
            var t = this;

            var undoFunc = function() {
                calendarEvent = calendarEvent._item;
                var calendarChange = false;
                if (calendarEvent._calendarId != calendarEvent._calendarIdOrig) {
                    calendarEvent._calendarId = calendarEvent._calendarIdOrig;
                    calendarChange = true;
                }
                calendarEvent._startTime = t._startTime;
                calendarEvent._endTime = t._endTime;
                if (calendar.store) {
                    calendar.store.put(calendarEvent);
                    if (calendarChange) {
                        calendar._refreshItemsRendering();
                    }
                }
            }

            serverRequest("UI.Utility","RoutineOpen",{modules:["OPS"],routine:"SYNC"}).then(function(moduleName) {
                if (moduleName !== "") {
                    undoFunc();
                    actionShowMessage("The Synchronization process must be complete before selecting this option");
                } else {
                    var changes = t.getChanges(calendarEvent,calendarStore);
                    if (changes.length > 0) {
                        t.getConflicts(calendarEvent).then(function(conflictResult) {
                            actionShowMessage("Please confirm the following change(s): <br>"+changes.join("<br>")+conflictResult,[
                                {label:"Save",callback:function() {
                                    var options = {
                                        id: t.eventId,
                                        oldCalendarId: calendarEvent._item._calendarIdOrig,
                                        newCalendarId: calendarEvent._item._calendarId,
                                        eventDate: locale.format(new Date(calendarEvent.startTime),{selector:"date",datePattern:"yyyy-MM-dd"}),
                                        eventTime: (calendarEvent._item._allDay ? null : locale.format(new Date(calendarEvent.startTime),{selector:"time",timePattern:"HH:mm:ss"})),
                                        duration:  (calendarEvent._item._allDay ? calendarEvent._item.event.duration : dateUtils.difference(new Date(calendarEvent.startTime),new Date(calendarEvent.endTime),"minute"))
                                    };
                                    serverRequest(t.className,"SaveCalendarDragAndDropHandler",options).then(function() {
                                        /*t._startTime = calendarEvent.startTime;
                                        t._endTime = calendarEvent.endTime;
                    
                                        if (calendarEvent._item._calendarId !== calendarEvent._item._calendarIdOrig) {
                                            var index = -1;
                                            arrayUtils.some(t.calendarIds,function(calendarId,i) {
                                                if (calendarId === calendarEvent._item._calendarIdOrig) {
                                                    index = i;
                                                    return true;
                                                }
                                                return false;
                                            });
                                            t.calendarIds.splice(index,1,calendarEvent._item._calendarId);
                                            calendarEvent._item._calendarIdOrig = calendarEvent._item._calendarId;
                                            calendarEvent._item.calendar = calendarStore.get(calendarEvent._item._calendarId);
                                        }
                    
                                        arrayUtils.forEach(t.calendarEventArray,function(calendarEvent) {
                                            calendarEvent._startTime = t._startTime;
                                            calendarEvent._endTime = t._endTime;
                                            var displayedCalendar = t.getCalendarForDisplay(calendarStore.get(calendarEvent._calendarId));
                                            (typeof displayedCalendar === "object" ? displayedCalendar.name : displayedCalendar)
                                            calendarEvent.primaryDetails = (typeof displayedCalendar === "object" ? displayedCalendar.name : displayedCalendar);
                                            if (calendar.store) {
                                                calendar.store.put(calendarEvent);
                                            }
                                        });*/
                                        topic.publish("poll");
                                    },function(error) {
                                        actionShowMessage(error);
                                        undoFunc();
                                    });
                                }},
                                {
                                    label:"Cancel",
                                    callback:undoFunc
                                }
                            ]);
                        });
                    };
                };
            });
        },

        getChanges: function(calendarEvent,calendarStore) {
            var t = this;
            var changes = [];
            var newStartTime = new Date(calendarEvent.startTime);
            var newEndTime = new Date(calendarEvent.endTime);
            var oldStartTime = new Date(t._startTime);
            var oldEndTime = new Date(t._endTime);
            if (dateUtils.compare(oldStartTime,newStartTime)) {
                // Start time changed
                var dateChanged = dateUtils.compare(oldStartTime,newStartTime,"date");
                var timeChanged = dateUtils.compare(oldStartTime,newStartTime,"time");
                if (dateChanged !== 0) {
                    changes.push("Date changed from "+locale.format(oldStartTime,{selector:"date",datePattern:"MM/dd/yyyy"})+" to "+locale.format(newStartTime,{selector:"date",datePattern:"MM/dd/yyyy"}));
                }
                if (timeChanged !== 0) {
                    changes.push("Time changed from "+locale.format(oldStartTime,{selector:"date",datePattern:"HHmm"})+" to "+locale.format(newStartTime,{selector:"date",datePattern:"HHmm"}));
                }
            } else if (!calendarEvent._item.event._allDay && dateUtils.compare(oldEndTime,newEndTime)) {
                // Duration changed
                changes.push("Duration changed from "+dateUtils.difference(oldStartTime,oldEndTime,"minute")+" minutes to "+dateUtils.difference(newStartTime,newEndTime,"minute")+" minutes");
            }
            if (calendarEvent._item._calendarIdOrig !== calendarEvent._item._calendarId) {
                changes.push("Calendar changed from "+calendarStore.get(calendarEvent._item._calendarIdOrig).name+" to "+calendarStore.get(calendarEvent._item._calendarId).name);
            }
            return changes;
        },

        getConflicts: function(calendarEvent) {
            var t = this;
            var deferred = new Deferred();
            if (!conflictCheckingEnabled == 1) {
                deferred.resolve([]);
            } else {
                var calendarIds = t.calendarIds;

                if (calendarEvent._item._calendarId !== calendarEvent._item._calendarIdOrig) {
                    var index = -1;
                    arrayUtils.some(calendarIds,function(calendarId,i) {
                        if (calendarId === calendarEvent._item._calendarIdOrig) {
                            index = i;
                            return true;
                        }
                        return false;
                    });
                    calendarIds.splice(index,1,calendarEvent._item._calendarId);
                }

                var event = {
                    id: t.id,
                    _allDay: t._allDay,
                    division: t.division,
                    _startTime: new Date(calendarEvent.startTime),
                    duration: (t._allDay ? t.duration : dateUtils.difference(new Date(calendarEvent.startTime),new Date(calendarEvent.endTime),"minute")),
                    calendars: t.calendarIds,
                    travelMinutes: t.travelMinutes,
                    modifiedStamp: t.modifiedStamp
                };

                
                require(["thornberry/schedule/ConflictChecker"],function(conflictChecker) {
                    var checker = new conflictChecker();
                    checker.checkEvents([event]).then(function(conflicts) {
                        if (conflicts.length > 0) {
                            deferred.resolve("<br><br><span style='text-decoration:underline'>Conflicts</span>:<br>"+checker.display(conflicts,false));
                        } else {
                            deferred.resolve("");
                        }
                    });
                })
            }
            return deferred.promise;
        },

        getDetails: function() {
            var t = this;
            var deferred = new Deferred();
            if (t.details !== null) {
                deferred.resolve(t.details);
            } else {
                serverRequest(t.className,"GetCalendarDetails",t.eventId).then(function(details) {
                    t.details = details;
                    deferred.resolve(details);
                },function() {
                    deferred.reject();
                });
            }
            return deferred.promise;
        },

        getColor: function(calendarId) {
            // Quick fix for User Conference
            if (!Event.prototype.calendarStore.get(calendarId)) {
                return "";
            }

            return Event.prototype.calendarStore.get(calendarId).color.toHex();
        },

        getCalendarIconClass: function(calendarId) {
            // Quick fix for User Conference
            if (!Event.prototype.calendarStore.get(calendarId)) {
                return "";
            }

            return Event.prototype.calendarStore.get(calendarId).getIconClass();;
        },

        getCalendarForDisplay: function(calendar) {
            return "";
        },

        getCalendarByType: function(type) {
            var t = this;
            var calendar = null;
            arrayUtils.some(t.calendarIds,function(calendarId) {
                var calendarItem = Event.prototype.calendarStore.get(calendarId);
                if (calendarItem.type === type) {
                    calendar = calendarItem;
                    return true;
                }
                return false;
            });
            return calendar;
        },

        getTypeInitial: function() {
            var initial = "";
            if (this.className === "Schedule.Calendar.Event.VolunteerActivity") {
                initial = "E";
            } else {
                switch(this.type) {
                    case "Visit" :
                        initial = "V";
                        break;
                    case "Agency" :
                        initial = "G";
                        break;
                    case "Time Off" :
                        initial = "T";
                        break;
                    case "Other" :
                        initial = "O";
                        break;
                    case "Meeting" :
                    case "Patient Care Meeting" :
                        initial = "M";
                        break;
                    case "Call" :
                    case "On-Call" :
                        initial = "C";
                        break;
                    case "Appointment" :
                        initial = "A";
                        break;
                    case "Patient Indicator" :
                        initial = "I";
                        break;
                    default : 
                        initial = (this.className === "Schedule.Calendar.Query.Client" ? "E" : ""); // Volunteer activity or holiday
                }
            }
            return initial;
        },

        getStatusClass: function(status) {
            var statusClass = "";
            switch(status || this.status) {
                case "Scheduled" :
                case "Ready" :
                    statusClass = "status-scheduled";
                    break;
                case "Missed" :
                    statusClass = "status-missed";
                    break;
                case "Cancelled" :
                    statusClass = "status-cancelled";
                    break;
                case "Hold" :
                case "OASIS Hold" :
                    statusClass = "status-hold";
                    break;
                case "Approved" :
                    statusClass = "status-approved";
                    break;
                case "Completed" :
                    statusClass = "status-completed";
                    break;
                case "Unauthorized" :
                    statusClass = "status-unauthorized";
                    break;
                case "Unassigned" :
                    statusClass = "status-unassigned";
                    break;
                case "Unapproved" :
                    statusClass = "status-unapproved";
                    break;
                case "Started" :
                case "In Progress" :
                    statusClass = "status-in-progress";
                    break;
                case "Not Synchronized" :
                    statusClass = "status-not-synced";
                    break;
                default : 
                    statusClass = "status-none";
            }
            return statusClass;
        },

        getActions: function() {
            var actions = [];
            if (this.editable) {
                actions.push("Edit");
            }
            return actions;
        },

        onAction: function(action,dateDefault,timeDefault) {},

        onCancelAction: function(action) {},

        openLink: function(link) {},

        getLinks: function() {
            return [];
        },

        getTypeIndexValue: function() {return "Event"}
    });

    var Visit = declare(Event,{
        className: "Schedule.Event.ScheduledVisit", // Default to scheduled visit for when a new visit is created

        getActions: function() {
            var t = this;
            var actions = this.inherited(arguments);
            if (t.other.paCancel == 0) {
                if (t.className === "Schedule.Event.ScheduledVisit") {
                    var dateDifference = dateUtils.difference(new Date(t._startTime),new Date(),"day");
                    if (EnvironmentSettings.userAccess.visitCharting == 1 && t.other.userId === userId && dateDifference <= 5 && dateDifference >= 0) {
                        actions.push("Chart");
                    }
                }
                if (this.editable && this.className === "Schedule.Event.ScheduledVisit") {
                    actions.push("Delete");
                }
            }
            return actions;
        },

        onAction: function(action,defaultDate,defaultTime,defaultCalendar) {
            var t = this;
            if (action === "Chart") {
                t.openLink({"id":1});
            } else if (action === "Edit" && typeof t.other.contactActivityId !== "undefined") {
                (new ContactActivity({eventId:t.other.contactActivityId})).onAction(action);
            } else if (action === "Edit" && typeof t.other.volunteerActivityId !== "undefined") {
                (new VolunteerActivity({eventId:t.other.volunteerActivityId})).onAction(action);
            } else {
                t.loadEditor(action,defaultDate,defaultTime,defaultCalendar).then(function(editor) {
                    Event.onLoadEditor(t,action,editor);
                },actionShowMessage);
            }
        },

        onCancelAction: function(action) {
            var t = this;
            if (action === "Edit") {
                if (typeof t.other.contactActivityId !== "undefined") {
                    (new ContactActivity({eventId:t.other.contactActivityId})).onCancelAction(action);
                } else if (typeof t.other.volunteerActivityId !== "undefined") {
                    (new VolunteerActivity({eventId:t.other.volunteerActivityId})).onCancelAction(action);
                } else {
                    var editorClass = (t.className === "Schedule.Event.ScheduledVisit" ? "Schedule.Calendar.Editor.ScheduledVisit" : "Schedule.Calendar.Editor.Visit");
                    when (t._loadEditor(editorClass),function(editor) {
                        editor.lock(t.eventId,true);
                    });
                }
            }
        },

        _loadEditor: function(editorClass) {
            if (typeof Visit.editor[editorClass] !== "undefined") {
                return Visit.editor[editorClass];
            } else {
                return load(editorClass,"Editor.Layout.CalendarVisit").then(function(editor) {
                    editor.startup();
                    Visit.editor[editorClass] = editor;
                    return editor;
                });
            }
        },

        loadEditor: function(action,defaultDate,defaultTime,defaultCalendar) {
            var t = this;
            var deferred = new Deferred();

            var editorClass = (t.className === "Schedule.Event.ScheduledVisit" ? "Schedule.Calendar.Editor.ScheduledVisit" : "Schedule.Calendar.Editor.Visit");

            when (t._loadEditor(editorClass),function(editor) {
                if (action === "Add") {
                    var patient = controller.getPatient();
                    editor.getPropertyDefinition("Patient").defaultValue = (patient !== null ? patient : "");
                    var employeeDefault = {id:"",name:""};
                    var calendars = controller.getSelectedCalendars();
                    if (defaultCalendar) {
                        arrayUtils.some(calendars,function(calendar) {
                            if (calendar.other.scheduledVisitEditable && calendar.id === defaultCalendar) {
                                employeeDefault = {
                                    id: calendar.other.userId,
                                    name: calendar.name+" "+calendar.identifier,
                                    empId: calendar.calendarId,
                                    teams: calendar.teams
                                }
                                return true;
                            }
                            return false;
                        });
                    }
                    if (employeeDefault.id === "") {
                        arrayUtils.some(calendars,function(calendar) {
                            if (calendar.type === "Employee" && calendar.other.scheduledVisitEditable) {
                                employeeDefault = {
                                    id: calendar.other.userId,
                                    name: calendar.name+" "+calendar.identifier,
                                    empId: calendar.calendarId,
                                    teams: calendar.teams
                                }
                                return true;
                            }
                            return false;
                        });
                    }
                    editor.getPropertyDefinition("Employee").defaultValue = employeeDefault;

                    editor.getPropertyDefinition("StartDate").defaultValue = (defaultDate ? locale.format(defaultDate,{selector:"date",datePattern:"yyyy-MM-dd"}) : "");
                    editor.getPropertyDefinition("StartTime").defaultValue = (defaultTime ? locale.format(defaultTime,{selector:"time",timePattern:"HH:mm:ss"}) : "");

                    editor.add();
                    deferred.resolve(editor);
                } else {
                    var onReject = function(error) {deferred.reject(error);}

                    editor.lock(t.eventId).then(function(success) {
                        if (success) {
                            editor.openById(t.eventId,action).then(function() {
                                deferred.resolve(editor);
                            },onReject);
                        } else {
                            onReject("This activity is being edited in another session");
                        }
                    },onReject);
                }
            });
            return deferred.promise;
        },

        getDetails: function() {
            var t = this;
            t.details = null; // Make details reload each time since changes to scheduling comment and clinical directives may have occurred
            return t.inherited(arguments);
        },

        getCalendarForDisplay: function(calendar) {
            var t = this;
            var displayCalendar = "";
            // Return the name of the assigned employee if a patient calendar and vice versa
            if (calendar.type === "Patient") {
                displayCalendar = t.getCalendarByType("Employee");
                if (!displayCalendar) {
                    if (t.other.auxEmployeeCalendar) {
                        displayCalendar = t.other.auxEmployeeCalendar;
                    } else if (t.other.employeeName) {
                        displayCalendar = t.other.employeeName;
                    } else {
                        displayCalendar = "Unassigned";
                    }
                }
            } else if (calendar.type === "Employee") {
                displayCalendar = t.getCalendarByType("Patient");
            }
            return displayCalendar;
        },

        openLink: function(link) {
            var t = this;
            var ndocNum = t.getCalendarByType("Patient").calendarId;
            if (link.id === 3) {
                startWindow.OpenVisitSummary(ndocNum,t.other.certNum,t.id);
            } else if (link.id === 4) {
                startWindow.OpenDashboardItem("ADMN","MODV","","MODV/"+ndocNum+"/"+t.id);
                window.focusOpen("","ADMNWINDOW");
            } else if (link.id === 2) {
                startWindow.OpenDashboardItem("OPS","VISAPR","","VisitApproval//"+ndocNum+"/"+t.id);
                window.focusOpen("","OPSWINDOW");
            } else if (link.id === 1) {
                if (t.className === "Schedule.Event.ScheduledVisit") {
                    startWindow.OpenDashboardItem("IN","VIS",ndocNum,"Visit/1/"+t.other.discipline+"/SCH:"+t.id);
                } else {
                    startWindow.OpenDashboardItem("IN","VIS",ndocNum,"Visit/1/"+t.other.discipline+"/"+t.id+"/TDY");
                }
                window.focusOpen("","INWINDOW");
            }
        },

        getLinks: function() {
            var t = this;
            var text = "";
            var linkArray = [];
            var dateDifference = dateUtils.difference(new Date(t._startTime),new Date(),"day");
            if (t.className === "Schedule.Event.ScheduledVisit") {
                if (EnvironmentSettings.userAccess.visitCharting == 1 && t.other.userId === userId && dateDifference <= 5 && dateDifference >= -5 && t.other.paCancel == 0) {
                    linkArray.push({id:1,label:"Chart this visit"});
                }
            } else if (t.className === "Schedule.Event.Visit") {
                linkArray.push({id:3,label:"Open Visit Summary"});
                if (t.status === "Hold" && EnvironmentSettings.userAccess.visitCharting == 1) {
                    linkArray.push({id:1,label:"Open in Visit Charting"});
                } else if (t.other.unapproved && EnvironmentSettings.userAccess.visitApproval == 1) {
                    linkArray.push({id:2,label:"Open Visit Approval"});
                }
                if (EnvironmentSettings.userAccess.modifyChartedVisit == 1 && dateDifference < parseInt(EnvironmentSettings.modifyPastVisits) && typeof t.other.volunteerActivityId === "undefined" && typeof t.other.contactActivityId === "undefined") {
                    linkArray.push({id:4,label:"Open in Modify Charted Visit",routine:"MODV",modules:["ADMN"]});
                }
            }
            return linkArray;
        },

        getTypeIndexValue: function() {return "Visit"}
    });
    Visit.editor = {};

    var Employee = declare(Event,{
        getActions: function() {
            var actions = this.inherited(arguments);
            if (this.editable) {
                actions.push("Delete");
            }
            return actions;
        },

        onAction: function(action,defaultDate,defaultTime,defaultCalendar) {
            var t = this;
            if (action === "Delete") {
                t.deleteEditor();
            } else {
                t.loadEditor(action,defaultDate,defaultTime,defaultCalendar).then(function(editor) {
                    Event.onLoadEditor(t,action,editor);
                },actionShowMessage);
            }
        },

        onCancelAction: function(action) {
            var t = this;
            if (action === "Edit") {
                when (t._loadEditor(),function(editor) {
                    editor.lock(t.eventId,true);
                });
            }
        },

        getEmployees: function() {
            var t = this;
            var employeeArray = [];
            var calendars = controller.getSelectedCalendars();
            arrayUtils.forEach(calendars,function(calendar) {
                if (calendar && calendar.type === "Employee" && t.allowEmployee(calendar)) {
                    employeeArray.push({
                        id: calendar.calendarId,
                        calendar: calendar.id,
                        name: calendar.name+" "+calendar.identifier
                    });
                }
            });
            return employeeArray;
        },

        loadEditor: function(action,defaultDate,defaultTime,defaultCalendar) {
            var t = this;
            var deferred = new Deferred();
            var employeeArray = [];
            var allowDefault = false;
            var employeeArray = t.getEmployees();
            if (action === "Add" && defaultCalendar) {
                allowDefault = arrayUtils.some(employeeArray,function(employeeItem) {
                    return employeeItem.calendar === defaultCalendar;
                });
            }

            onReject = function(error) {deferred.reject(error);}

            when(t._loadEditor(),function(editor) {
                editor.getPropertyDefinition("Employee").options = employeeArray;
                if (action === "Edit") {
                    editor.lock(t.eventId).then(function(success) {
                        if (success) {
                            editor.openById(t.eventId,action).then(function() {
                                deferred.resolve(editor);
                            },onReject);
                        } else {
                            onReject("This activity is being edited in another session");
                        }
                    },onReject);
                } else if (action === "Add") {
                    editor.getPropertyDefinition("StartDate").defaultValue = (defaultDate ? locale.format(defaultDate,{selector:"date",datePattern:"yyyy-MM-dd"}) : "");
                    editor.getPropertyDefinition("StartTime").defaultValue = (defaultTime ? locale.format(defaultTime,{selector:"time",timePattern:"HH:mm:ss"}) : "");

                    if (defaultCalendar && allowDefault) {
                        var calendarId = defaultCalendar.toString().split("||")[0];
                        editor.getPropertyDefinition("Employee").defaultValue = calendarId;
                    } else {
                        editor.getPropertyDefinition("Employee").defaultValue = "";
                    }

                    editor.add();
                    deferred.resolve(editor);
                }
            });
            return deferred.promise;
        },

        deleteEditor: function(editor) {
            var t = this;
            var deleteFunction = function() {
                when(t._loadEditor(),function(editor) {
                    editor.deleteEditor(t.eventId).then(function() {},actionShowMessage);
                    topic.publish("poll");
                });
            }

            actionShowMessage("Are you sure you want to delete this event?",[
                {label:"Yes",callback:deleteFunction},
                {label:"No"}
            ]);
        }
    });

    var Agency = declare(Employee,{
        allowEmployee: function(empCalendar) {
            return (empCalendar.other.agencyEditable == 1);
        },

        _loadEditor: function() {
            if (Agency.editor !== null) {
                return Agency.editor;
            } else {
                return load("Schedule.Calendar.Editor.Agency","Editor.Layout.CalendarAgency").then(function(editor) {
                    editor.startup();
                    Agency.editor = editor;
                    return editor;
                });
            }
        },

        getTypeIndexValue: function() {return "Employee:Agency"}
    });
    Agency.editor = null;

    var TimeOff = declare(Employee,{
        allowEmployee: function(empCalendar) {
            return (empCalendar.other.timeOffEditable == 1);
        },

        _loadEditor: function() {
            if (TimeOff.editor !== null) {
                return TimeOff.editor;
            } else {
                return load("Schedule.Calendar.Editor.TimeOff","Editor.Layout.CalendarTimeOff").then(function(editor) {
                    editor.startup();
                    TimeOff.editor = editor;
                    return editor;
                });
            }
        },

        getTypeIndexValue: function() {return "Employee:Time Off"}
    });
    TimeOff.editor = null;

    var EmployeeOther = declare(Employee,{
        allowEmployee: function(empCalendar) {
            return (empCalendar.other.otherEditable == 1);
        },

        _loadEditor: function() {
            if (EmployeeOther.editor !== null) {
                return EmployeeOther.editor;
            } else {
                return load("Schedule.Calendar.Editor.EmployeeOther","Editor.Layout.CalendarEmployeeOther").then(function(editor) {
                    editor.startup();
                    EmployeeOther.editor = editor;
                    return editor;
                });
            }
        },

        getTypeIndexValue: function() {return "Employee:Other"}
    });
    EmployeeOther.editor = null;

    var PatientActivity = declare(Event,{
        getActions: function() {
            var actions = this.inherited(arguments);
            if (this.other.enablePrepare) {
                actions.push("Prepare");
            }
            if (this.editable) {
                actions.push("Delete");
            }
            return actions;
        },

        onAction: function(action,defaultDate,defaultTime) {
            var t = this;
            if (action === "Delete") {
                t.deleteEditor();
            } else {
                var canContinue = true;
                if (action === "Edit") {
                    canContinue = serverRequest("Patient.Activity.Utility","LockActivity",{activityId:t.eventId,release:false});
                }
                when(canContinue,function() {
                    t.loadEditor(action,defaultDate,defaultTime).then(function(editor) {
                        Event.onLoadEditor(t,action,editor);
                    },actionShowMessage);
                },function(error) {
					actionShowMessage((error == 1 ? "There was an error acquiring a lock on this activity" : error));
				});
            }
        },

        onCancelAction: function(action) {
            var t = this;
            if (action === "Edit") {
                serverRequest("Patient.Activity.Utility","LockActivity",{activityId:t.eventId,release:true});
            }
        },

        loadEditor: function(action,defaultDate,defaultTime) {
            var t = this;

            var deferred = new Deferred();
            if (PatientActivity.editor) {
                PatientActivity.editor.destroyRecursive();
            }

            var activityTypeId = "";
            if (action === "Add") {
                if (t.type === "Meeting") {
                    activityTypeId = 3;
                } else if (t.type === "Call") {
                    activityTypeId = 1;
                } else if (t.type === "On-Call") {
                    activityTypeId = 2;
                } else if (t.type === "Other") {
                    activityTypeId = 5;
                }
            }

            require(["thornberry/patientActivity/ActivityEditor"],function(PatientActivityEditor){
                PatientActivityEditor.loadActivity(t.eventId,action,activityTypeId).then(function(editor) {
                    PatientActivity.editor = editor;
                    domStyle.set(editor.scrollableDiv,"overflow-y",""); // Allow outer div to control scrolling

                    if (action === "Add") {
                        if (defaultDate) {
                            editor.dateField.set("value",defaultDate);
                        }
                        if (defaultTime) {
                            editor.startTimeField.set("value",defaultDate);
                        }
                    }

                    aspect.around(editor,"save",function(saveFunction) {
                        return function() {
                            var deferred = new Deferred();
                            var returnObject = {
                                status: false,
                                code: "",
                                id: ""
                            };

                            if (editor.okToSave()) {
                                lang.hitch(editor,saveFunction)().then(function(id) {
                                    returnObject.status = true;
                                    returnObject.id = id;
                                    deferred.resolve(returnObject);
                                },function() {
                                    deferred.resolve(returnObject);
                                });
                            } else {
                                deferred.resolve(returnObject);
                            }
                            return deferred.promise;
                        };
                    });
                    deferred.resolve(editor);
                });
            });
            return deferred.promise;
        },

        getCalendarForDisplay: function(calendar) {
            var t = this;
            var displayCalendar = "";
            if (calendar.type !== "Patient") {
                displayCalendar = t.getCalendarByType("Patient");
            }
            return displayCalendar;
        },

        deleteEditor: function() {
            var t = this;
            var deleteFunction = function() {
                serverRequest("Patient.Activity.Utility","DeleteActivity",{
                    userId: userId,
                    deviceId: deviceId,
                    activityId: t.eventId
                }).then(function() {topic.publish("poll");},actionShowMessage);
            }

            actionShowMessage("Are you sure you want to delete this event?",[
                {label:"Yes",callback:deleteFunction},
                {label:"No"}
            ]);
        },

        openLink: function(link) {
            var t = this;
            var ndocNum = t.getCalendarByType("Patient").calendarId
            if (app.selectAccount(ndocNum)) {
                var modules = ["IN","OPS","ADMN"];
                serverRequest("UI.Utility","RoutineOpen",{modules:modules,routine:"PACTIVITY"}).then(function(module) {
                    if (module !== "") {
                        startWindow.window[module+"WINDOW"].MAIN.autoStart(t.eventId);
                        window.focusOpen('',module+'WINDOW').focus()
                    } else {
                        serverRequest("UI.Utility","SetAutoStartOption","PatientActivity/"+t.id).then(function() {
                            var module = startWindow.OpenRoutine(["IN","OPS","ADMN"],"PACTIVITY");
                            window.focusOpen('',module+'WINDOW').focus();
                        });
                    }
                });
            }
        },

        getLinks: function() {
            return (this.other.userHasAccess == 1 ? [{label:"Open in Patient Activity"}] : []);
        },

        getTypeIndexValue: function() {return "Patient:"+this.type;}
    });
    PatientActivity.editor = null;

    var ContactActivity = declare(Event,{
        className: "Schedule.Calendar.Event.ContactActivity",
        type: "Other",

        getActions: function() {
            var actions = this.inherited(arguments);
            if (this.editable) {
                actions.push("Delete");
            }
            return actions;
        },

        onAction: function(action,defaultDate,defaultTime) {
            var t = this;
            if (action === "Delete") {
                t.deleteEditor();
            } else {
                var canContinue = true;
                if (action === "Edit") {
                    canContinue = serverRequest("Contact.Activity.Utility","LockActivity",{activityId:t.eventId,release:false});
                }
                when(canContinue,function() {
                    t.loadEditor(action,defaultDate,defaultTime).then(function(editor) {
                        Event.onLoadEditor(t,action,editor);
                    },actionShowMessage);
                },function(error) {
					actionShowMessage((error == 1 ? "There was an error acquiring a lock on this activity" : error));
				});
            }
        },

        onCancelAction: function(action) {
            var t = this;
            if (action === "Edit") {
                serverRequest("Contact.Activity.Utility","LockActivity",{activityId:t.eventId,release:true});
            }
        },

        loadEditor: function(action,defaultDate,defaultTime) {
            var t = this;

            var deferred = new Deferred();
            if (ContactActivity.editor) {
                ContactActivity.editor.destroyRecursive();
            }

            var activityTypeId = "";
            if (action === "Add") {
                if (t.type === "Meeting") {
                    activityTypeId = 6;
                } else if (t.type === "Call") {
                    activityTypeId = 3;
                } else if (t.type === "Other") {
                    activityTypeId = 7;
                }
            }

            require(["thornberry/contactActivity/ActivityEditor"],function(ContactActivityEditor){
                ContactActivityEditor.loadActivity(t.eventId,action,activityTypeId).then(function(editor) {
                    ContactActivity.editor = editor;
                    domStyle.set(editor.scrollableDiv,"overflow-y",""); // Allow outer div to control scrolling

                    if (action === "Add") {
                        if (defaultDate) {
                            editor.dateField.set("value",defaultDate);
                        }
                        if (defaultTime) {
                            (editor.startTimeField ? editor.startTimeField : editor.timeField).set("value",defaultDate);
                        }
                    }

                    aspect.around(editor,"save",function(saveFunction) {
                        return function() {
                            var deferred = new Deferred();
                            var returnObject = {
                                status: false,
                                code: "",
                                id: ""
                            };

                            if (editor.okToSave()) {
                                lang.hitch(editor,saveFunction)().then(function(id) {
                                    returnObject.status = true;
                                    returnObject.id = id;
                                    deferred.resolve(returnObject);
                                },function() {
                                    deferred.resolve(returnObject);
                                });
                            } else {
                                deferred.resolve(returnObject);
                            }
                            return deferred.promise;
                        };
                    });
                    deferred.resolve(editor);
                });
            });
            return deferred.promise;
        },

        getCalendarForDisplay: function(calendar) {
            var t = this;
            var displayCalendar = "";
            if (calendar.type !== "Contact") {
                displayCalendar = t.getCalendarByType("Contact");
            }
            return displayCalendar;
        },

        deleteEditor: function() {
            var t = this;
            var deleteFunction = function() {
                serverRequest("Contact.Activity.Utility","DeleteActivity",{
                    userId: userId,
                    deviceId: deviceId,
                    activityId: t.eventId
                }).then(function() {topic.publish("poll");},actionShowMessage);
            }

            actionShowMessage("Are you sure you want to delete this event?",[
                {label:"Yes",callback:deleteFunction},
                {label:"No"}
            ]);
        },

        openLink: function(link) {
            var t = this;
            var ndocNum = t.getCalendarByType("Contact").calendarId;
            if (app.selectAccount(ndocNum)) {
                var modules = ["OPS"];
                serverRequest("UI.Utility","RoutineOpen",{modules:modules,routine:"CACT"}).then(function(module) {
                    if (module !== "") {
                        startWindow.window[module+"WINDOW"].MAIN.autoStart(t.eventId);
                        window.focusOpen('','OPSWINDOW').focus()
                    } else {
                        startWindow.OpenDashboardItem("OPS","CACT",ndocNum,"ContactActivity/"+t.id);
                        window.focusOpen('','OPSWINDOW').focus()
                    }
                });
            }
        },

        getLinks: function() {
            return (this.other.userHasAccess == 1 ? [{label:"Open in Contact Activity"}] : []);
        },

        getTypeIndexValue: function() {return "Contact:"+this.type;}
    });
    ContactActivity.editor = null;

    var VolunteerActivity = declare(Event,{
        className: "Schedule.Calendar.Event.VolunteerActivity",
        type: "Volunteer Activity",

        getActions: function() {
            var actions = this.inherited(arguments);
            if (this.editable) {
                actions.push("Copy");
                actions.push("Error");
            }
            return actions;
        },

        onAction: function(action,defaultDate,defaultTime,defaultCalendar) {
            var t = this;
            var canContinue = true;
            if (action === "Edit" || action === "Error") {
                canContinue = serverRequest("Volunteer.Utility","LockActivity",{activityId:t.eventId,release:false});
            }
            when(canContinue,function() {
                t.loadEditor(action,defaultDate,defaultTime,defaultCalendar).then(function(editor) {
                    Event.onLoadEditor(t,action,editor);
                    domStyle.set(editor.containerNode,"height","100%");
                    editor.resize();
                },actionShowMessage);
            },function(error) {
                actionShowMessage((error == 1 ? "There was an error acquiring a lock on this activity" : error));
            });
        },

        onCancelAction: function(action) {
            var t = this;
            if (action === "Edit" || action === "Error") {
                serverRequest("Volunteer.Utility","LockActivity",{activityId:t.eventId,release:true});
            }
        },

        loadEditor: function(action,defaultDate,defaultTime,defaultCalendar) {
            var t = this;

            var deferred = new Deferred();
            require(["thornberry/volunteer/ActivityEditor","thornberry/volunteer/utils"],function(VolunteerActivityEditor,volunteerUtils) {
                VolunteerActivityEditor.loadActivity(t.eventId,action).then(function(editor) {
                    if (action === "Add") {
                        var volunteerCalendar = null;
                        var calendarStore = new Memory({data:controller.getSelectedCalendars()});

                        if (defaultCalendar) {
                            var calendar = calendarStore.get(defaultCalendar);
                            if (calendar && calendar.other.volunteer) {
                                volunteerCalendar = calendar; 
                            }
                        }
                        if (!volunteerCalendar) {
                            arrayUtils.some(calendarStore.query({},{sort:[{attribute:"name"}]}),function(calendar) {
                                if (calendar.other.volunteer) {
                                    volunteerCalendar = calendar;
                                    return true;
                                }
                                return false;
                            });
                        }
                        if (volunteerCalendar) {
                            editor.addVolunteer(volunteerCalendar.calendarId);
                        }

                        if (defaultDate) {
                            editor.startDateField.set("value",defaultDate);
                        }
                        if (defaultTime) {
                            editor.startTimeField.set("value",defaultTime);
                        }
                    }
                    if (action === "Error") {
                        editor.defer(function() {
                            editor.set("modified",true);
                        });
                    }
                    if (!editor.save) {
                        editor.save = function() {
                            var deferred = new Deferred();
                            var returnObject = {
                                status: false,
                                code: "",
                                id: ""
                            };
                            
                            this.okToSave().then(function() {
                                device = deviceId;
                                volunteerUtils.serverManager.saveActivities(editor.action,editor.getActivityArray()).then(function() {
                                    delete device;
                                    returnObject.status = true;
                                    deferred.resolve(returnObject);
                                },function() {
                                    deferred.resolve(returnObject);
                                });
                            },function() {
                                deferred.resolve(returnObject);
                            });
                            return deferred.promise;
                        };

                        editor.saveAndAdd = function() {
                            var t = this;
                            var deferred = new Deferred();
                            if (t.action === "Add") {
                                t.save().then(function(returnObject) {
                                    if (returnObject.status) {
                                        var object = editor.getActivityArray()[0];
                                        object.endDate = editor.activity.endDate;
                                        object.endTime = editor.activity.endTime;
                                        object.totalTime = editor.activity.totalTime;
                                        t.edit("Add",object);
                                    }
                                    deferred.resolve(returnObject);
                                });
                            }
                            return deferred.promise;
                        };
                    }

                    domStyle.set(editor.titleContainerNode,"margin","0px"); // Allow horizontal scrollbar to display
                    domStyle.set(editor.domNode,"overflow","hidden");

                    deferred.resolve(editor);
                });
            });
            return deferred.promise;
        },

        getCalendarForDisplay: function(calendar) {
            var t = this;
            var displayCalendar = "";
            if (calendar.type !== "Employee") {
                displayCalendar = t.getCalendarByType("Employee");
            } else {
                displayCalendar = t.getCalendarByType("Patient");
                if (!displayCalendar) {
                    displayCalendar = t.other.patientCalendar || "";
                }
            }
            return displayCalendar;
        },

        openLink: function(link) {
            var t = this;
            var employeeId = t.getCalendarByType("Employee").calendarId;
            serverRequest("UI.Utility","RoutineOpen",{modules:["OPS"],routine:"VolunteerActivity"}).then(function(module) {
                if (module !== "") {
                    startWindow.window[module+"WINDOW"].MAIN.autoStart(employeeId,t.eventId);
                    window.focusOpen('','OPSWINDOW').focus()
                } else {
                    startWindow.OpenDashboardItem("OPS","VolunteerActivity","","VolunteerActivity/"+employeeId+"/"+t.eventId);
                    window.focusOpen('','OPSWINDOW').focus()
                }
            });
        },

        getLinks: function() {
            return (this.other.userHasAccess == 1 ? [{label:"Open Volunteer Activity"}] : []);
        },
        
        getTypeIndexValue: function() {return "Volunteer Activity"}
    });

    var PatientCareMeeting = declare(Event,{
        onAction: function(action,defaultDate,defaultTime) {
            var t = this;
            var deferred = new Deferred();

            t.openLink(action,defaultDate,defaultTime);

            if (action === "Add") {
                setTimeout(function() {
                    controller.popView();
                },1000);
                deferred.resolve();
            }

            return deferred.promise;
        },

        openLink: function(link,defaultDate,defaultTime) {
            var t = this;
            var modules = ["IN","OPS","ADMN"];
            serverRequest("UI.Utility","RoutineOpen",{modules:modules,routine:"PatientCareMeetings"}).then(function(module) {
                defaultDate = (defaultDate ? locale.format(defaultDate,{selector:"date",datePattern:"yyyy-MM-dd"}) : "");
                defaultTime = (defaultTime ? locale.format(defaultTime,{selector:"time",timePattern:"HHmm"}) : "");

                if (module !== "") {
                    if (link === "Add") {
                        startWindow.window[module+"WINDOW"].MAIN.autoStart("Add",defaultDate,defaultTime);
                    } else {
                        startWindow.window[module+"WINDOW"].MAIN.autoStart(t.eventId,t.other.scheduledDate);
                    }
                    window.focusOpen('',module+'WINDOW').focus()
                } else {
                    if (link === "Add") {
                        var autoStartOptions = "PatientCareMeetings/Add/"+defaultDate+"/"+defaultTime;
                    } else {
                        var autoStartOptions = "PatientCareMeetings/"+t.eventId+"/"+t.other.scheduledDate;
                    }

                    serverRequest("UI.Utility","SetAutoStartOption",autoStartOptions).then(function() {
                        var module = startWindow.OpenRoutine(modules,"PatientCareMeetings");
                        window.focusOpen('',module+'WINDOW').focus();
                    });
                }
            });
        },

        getLinks: function() {
            return (this.other.userHasAccess == 1 ? [{label:"Open in Patient Care Meetings"}] : []);
        },
        
        getTypeIndexValue: function() {return "Patient Care Meeting"}
    });
    PatientCareMeeting.editor = null;

    var Appointment = declare(Event,{
        getActions: function() {
            var actions = this.inherited(arguments);
            if (this.editable) {
                actions.push("Delete");
            }
            return actions;
        },

        onAction: function(action,defaultDate,defaultTime) {
            var t = this;
            if (action === "Delete") {
                t.deleteEditor();
            } else {
                t.loadEditor(action,defaultDate,defaultTime).then(function(editor) {
                    Event.onLoadEditor(t,action,editor);
                },actionShowMessage);
            }
        },

        onCancelAction: function(action) {
            var t = this;
            if (action === "Edit") {
                when (t._loadEditor(),function(editor) {
                    editor.lock(t.eventId,true);
                });
            }
        },

        _loadEditor: function(editorClass) {
            if (Appointment.editor) {
                return Appointment.editor;
            } else {
                return load("Schedule.Calendar.Editor.Appointment","Editor.Layout.CalendarAppointment").then(function(editor) {
                    editor.startup();
                    Appointment.editor = editor;
                    return editor;
                });
            }
        },

        loadEditor: function(action,defaultDate,defaultTime) {
            var t = this;
            var deferred = new Deferred();

            onReject = function(error) {deferred.reject(error);}

            when (t._loadEditor(),function(editor) {
                if (action === "Edit") {
                    editor.lock(t.eventId,false,t.calendarIds[0]).then(function(success) {
                        if (success) {
                            editor.openById(t.eventId,action).then(function() {
                                deferred.resolve(editor);
                            },onReject);
                        } else {
                            onReject("This activity is being edited in another session");
                        }
                    },onReject);
                } else if (action === "Add") {
                    var patient = controller.getPatient();
                    editor.getPropertyDefinition("NDocNumber").defaultValue = (patient !== null ? patient.value : "");
                    editor.getPropertyDefinition("StartDate").defaultValue = (defaultDate ? locale.format(defaultDate,{selector:"date",datePattern:"yyyy-MM-dd"}) : "");
                    editor.getPropertyDefinition("StartTime").defaultValue = (defaultTime ? locale.format(defaultTime,{selector:"time",timePattern:"HH:mm:ss"}) : "");
                    editor.add();
                    deferred.resolve(editor);
                }
            });
            return deferred.promise;
        },

        deleteEditor: function(editor) {
            var t = this;

            var deleteFunction = function() {
                when(t._loadEditor(),function(editor) {
                    editor.deleteEditor(t.eventId).then(function() {topic.publish("poll");},actionShowMessage);
                });
            }

            actionShowMessage("Are you sure you want to delete this event?",[
                {label:"Yes",callback:deleteFunction},
                {label:"No"}
            ]);
        },
        
        getTypeIndexValue: function() {return "Appointment"}
    });
    Appointment.editor = null;

    var PatientIndicator = declare(Event,{
		getLupaWarning: function() {
            var t = this;
			var value = false;
            if (t.other.periodType === "billingPeriod") {
				if (t.other.lupaWarning == 1 || t.other.lupaWarningSch == 1) {
					var filterEditor = registry.byId("calendarFilterEditor");
					value = (filterEditor && filterEditor.object.LupaType === "scheduled" ? t.other.lupaWarningSch == 1 : t.other.lupaWarning == 1);
				}
			}
			return value;
        },
		
        getTypeIndexValue: function() {return "Patient Indicator"}
    });

	Event.fixAllDayEvent = function(event) {
		//CLR added 2/17/2020
		//fix for Jira ND-9740
		//Note that this change is to correct a bug within the Dojo calendar widget
		//and should not be carried into the UI project.
		
		//validate event times first
		if (event._startTime && event._endTime) {
		
			//get the start datetime values into Date objects
			var sDate = new Date(event._startTime);
			var eDate = new Date(event._endTime);

			//if the start and end date are the same, and the user entered start and stop times of 00:00,
			//then we'll need to tweak some values or the event won't show on the calendar.
			if (sDate.toDateString() === eDate.toDateString() && 
				sDate.getHours() === 0 && sDate.getMinutes() === 0 && 
				eDate.getHours() === 0 && eDate.getMinutes() === 0) {
				
				//set ending time to 23:59
				eDate.setHours(23);
				eDate.setMinutes(59);
				
				//convert back to a string and flag as an all day event
				event._endTime = eDate.toISOString();
				event._allDay = 1;
				
				//this one is questionable.  By default, event.time is set to "00:00:00".  To make it match other all day events,
				//though, I'm clearing the value.  This causes the event details to show "0 minutes" under the date of the event.
				//It's nicer with this set to 00:00:00 because the calendar doesn't show anything under the date of the event, but
				//for consistency, I am clearing it anyway.
				event.time = "";
			}
		}
		
		return event;
		
	}

    Event.constructEvent = function(data) {
        var event = null;
        if (data.className === "Schedule.Calendar.Event.PatientActivity") {
            event = new PatientActivity(data)
        } else if (data.className === "Schedule.Calendar.Event.ContactActivity") {
            event = new ContactActivity(data)
        } else if (data.className === "Schedule.Calendar.Event.VolunteerActivity") {
            event = new VolunteerActivity(data)
        } else {
            switch(data.type) {
                case "Visit":
                    event = new Visit(data)
                    break;
                case "Agency":
                    event = new Agency(data)
                    break;
                case "Time Off":
                    event = new TimeOff(data)
                    break;
                case "Other":
                    event = new EmployeeOther(data)
                    break;
                case "Patient Care Meeting":
                    event = new PatientCareMeeting(data)
                    break;
                case "Appointment":
                    event = new Appointment(data)
                    break;
                case "Patient Indicator":
                    event = new PatientIndicator(data)
                    break;
                default:
                    event = new Event(data);
            }
        }
		
        return Event.fixAllDayEvent(event);
    };

    Event.userProperties = {}; // For user access/other values gathered from the server

    Event.onLoadEditor = function(event,actionType,editor) {};

    return Event;
});