// thornberry/editor/Editor.js
define([
	"dojo/_base/declare",
	"dijit/_WidgetBase",
	"dijit/_TemplatedMixin",
    "dijit/_WidgetsInTemplateMixin",
    "dojo/Deferred",
    "dojo/promise/all",
    "thornberry/serverRequest",
    "dojo/_base/array",
    "dojo/_base/lang",
    "dojo/on",
    "dojo/when",
    "dojo/topic",
    "dojo/Stateful",
    "dojo/query",
    "dojo/json",
    "thornberry/object/TableManager"
], function(declare,_WidgetBase,_TemplatedMixin,_WidgetsInTemplateMixin,Deferred,all,serverRequest,arrayUtils,lang,on,when,topic,Stateful,query,JSON,TableManager){
    var Editor = declare([_WidgetBase, _TemplatedMixin, _WidgetsInTemplateMixin], {
        baseClass: "thornberryEditor",
        templateString: "",
        editorClass: "",
        definition: null,
        settings: null,
        scratch: null, // provided for arbitrary use by developers
        parentEditor: null,
        saveAndAddAllowed: true,

        disabled: false,

        action: "",
        initialObject: null,
        modifiedObject: null,

        fields: null,
        
        _onValidates: null, // hash of functions
        _hasServerOnValidates: null, // hash of booleans
        _onChanges: null, // hash of functions
        _hasServerOnChanges: null, // hash of booleans
        _isRequireds: null, // hash of functions
        _isDisableds: null, // hash of functions
        _modified: false,
        _messages: null, // array of messages to be displayed to the end user
        _lastValue: null,

        _watcherHandle: null,

        onActionDeferred: null, // created each time an action (i.e., add, edit, etc.) is started, and resolves when the editor is ready
        loadDeferred: null,
        _loadPromiseArray: null,

        postMixInProperties: function() {
            var t = this;
            t.inherited(arguments);

            t.scratch = {};

            if (t.definition) {
                t.settings = t.definition.settings;
            }

            t.fields = {};

            t._onValidates = {};
            t._hasServerOnValidates = {};
            t._onChanges = {};
            t._hasServerOnChanges = {};
            t._isRequireds = {};
            t._isDisableds = {};
            t._messages = [];

            t.loadDeferred = new Deferred();
            t._loadPromiseArray = [];

            if (t.definition.onLoad) {
                t.loadDeferred.then(function() {
                    t._onLoad();
                });
            }
        },

        buildRendering: function() {
            var t = this;
            t.inherited(arguments);
			
            // First, load all modules
            if (t.definition) {
                var moduleDeferred = new Deferred();
                var tableDeferred = new Deferred();
                t._loadPromiseArray.push(moduleDeferred.promise);
                t._loadPromiseArray.push(tableDeferred.promise);
                var modulePathArray = arrayUtils.map(t.definition.moduleDefinitions,function(definition) {return definition.path});
                require(modulePathArray,function() {
                    for (var i = 0; i < arguments.length; i++) {
                        eval("var "+t.definition.moduleDefinitions[i].name+" = arguments[i];");
                    }

                    if (t.definition.onChange) {
                        eval("var onChangeFunction = "+t.definition.onChange);
                        t._onChange = onChangeFunction;
                    }

                    if (t.definition.onLoad) {
                        eval("var onLoadFunction = "+t.definition.onLoad);
                        t._onLoad = onLoadFunction;
                    }

                    if (t.definition.getObjectActions) {
                        eval("var getObjectActionsFunction = "+t.definition.getObjectActions);
                        t.getObjectActions = getObjectActionsFunction;
                    }

                    if (t.definition.onAction) {
                        eval("var onActionFunction = "+t.definition.onAction);
                        t.onAction = onActionFunction;
                    }

                    if (t.definition.onValidate) {
                        eval("var onValidateFunction = "+t.definition.onValidate);
                        t._onValidate = onValidateFunction;
                    }

                    var loadPropertyArray = [];
                    arrayUtils.forEach((t.definition.propertyDefinitions || []), function(propertyDefinition) {
                        var deferred = new Deferred();
                        loadPropertyArray.push(deferred.promise);

                        // TODO: use propertyDefinition.collectionWidget here instead if this property is a collection property

                        require([propertyDefinition.widget],function(Widget) {
                            try {
                                var field = new Widget({
                                    editor: t,
                                    propertyDefinition: propertyDefinition,
                                    constraints: propertyDefinition.constraints
                                },t[propertyDefinition.name+"Node"]);
                                field.startup();

                                if (propertyDefinition.required) {
                                    field.set("requiredProperty",true);
                                }
                                if (propertyDefinition.disabled) {
                                    field.set("disabled",true);
                                }

                                t.own(on(field,"change",function() {t._modifyValue(this.propertyDefinition.name)}));

                                if (propertyDefinition.onValidate) {
                                    eval("var onValidateFunction = "+propertyDefinition.onValidate);
                                    t._onValidates[propertyDefinition.name] = onValidateFunction;
                                }
                                if (propertyDefinition.hasServerOnValidate) {
                                    t._hasServerOnValidates[propertyDefinition.name] = true;
                                }
                                if (propertyDefinition.onChange) {
                                    eval("var onChangeFunction = "+propertyDefinition.onChange);
                                    t._onChanges[propertyDefinition.name] = onChangeFunction;
                                }
                                if (propertyDefinition.hasServerOnChange) {
                                    t._hasServerOnChanges[propertyDefinition.name] = true;
                                }
                                if (propertyDefinition.isDisabled && !propertyDefinition.disabled) {
                                    eval("var isDisabledFunction = "+propertyDefinition.isDisabled);
                                    t._isDisableds[propertyDefinition.name] = isDisabledFunction;
                                }
                                if (propertyDefinition.isRequired && !propertyDefinition.required) {
                                    eval("var isRequiredFunction = "+propertyDefinition.isRequired);
                                    t._isRequireds[propertyDefinition.name] = isRequiredFunction;
                                }

                                if (field._loadDeferred) {
                                    // Wait until the widget has completed any asynchronous loading
                                    t._loadPromiseArray.push(field._loadDeferred.promise);
                                }

                                t.fields[propertyDefinition.name] = field;
                                deferred.resolve();
                            } catch (e) {
                                deferred.reject(e);
                            }
                        });
                    });
                    var propertyPromise = all(loadPropertyArray);
                    
                    propertyPromise.then(function() {
                        // Load any queued tables
                        TableManager.load().then(function() {
                            tableDeferred.resolve();
                        });
                    });
                    t._loadPromiseArray.push(propertyPromise);
                    moduleDeferred.resolve();
                });
            }
        },

        postCreate: function() {
            var t = this;
            t.inherited(arguments);

            t._loadPromiseArray.push(t.loadSettings());

            if (t._loadPromiseArray.length > 0) {
                all(t._loadPromiseArray).then(function() {
                    // Check again for any promises that may have been added after the first "all" call
                    all(t._loadPromiseArray).then(function() {
                        t.loadDeferred.resolve();
                    });
                });
            } else {
                t.loadDeferred.resolve();
            }
        },

        loadSettings: function() {
            var t = this;
            return serverRequest(t.editorClass,"LoadSettings",{}).then(function(settings) {
                t.settings = settings;
            });
        },

        lock: function(id,release,ndocNum) {
            var t = this;
            if (typeof release === "undefined") {
                release = false;
            }
            ndocNum = ndocNum || "";
            return serverRequest("Editor.Loader","Lock",{editorClass: t.editorClass,id:id,release:release,ndocNum:ndocNum});
        },

        openById: function(id,action) {
            var t = this;
            t.action = action;
            t.onActionDeferred = new Deferred();
            return serverRequest("Editor.Loader","Read",{editorClass:t.editorClass,id:id,action:action}).then(function(object) {
                t.open(object,action);
            });
        },
		
		add: function() {
            var t = this;
            t.onActionDeferred = new Deferred();
			t.open(t.getNewObject(),"Add");
        },
        
        getNewObject: function() {
            var t = this;
            var object = {};
			if (t.definition) {
                arrayUtils.forEach((t.definition.propertyDefinitions || []), function(propertyDefinition) {
					object[propertyDefinition.name] = propertyDefinition.defaultValue;
                });
            }
            return object;
        },

        open: function(object,action,initialObject) {
            var t = this;
            t.action = action;
            if (!t.onActionDeferred) {
                t.onActionDeferred = new Deferred();
            }

            if (t._watcherHandle) {
                t._watcherHandle.remove();
            }
            
            if (object) {
                t.initialObject = (typeof initialObject !== "undefined" ? initialObject : object || {});
                t.modifiedObject = new Stateful(lang.clone(object || {}));
                t._watcherHandle = t.modifiedObject.watch(lang.hitch(t,"_notify"));
            }

            if (t.definition) {
                arrayUtils.forEach((t.definition.propertyDefinitions || []), function(propertyDefinition) {
                    var serializedValue = (object ? object[propertyDefinition.name] : propertyDefinition.defaultValue);
                    t.fields[propertyDefinition.name].deserialize(serializedValue,true);
                });

                // After all initial values have been set, call onChangeProperty() when defined
                arrayUtils.forEach((t.definition.propertyDefinitions || []), function(propertyDefinition) {
                    var onChange = t._onChanges[propertyDefinition.name];
                    if (typeof onChange === "function") {
                        var serializedValue = (object ? object[propertyDefinition.name] : propertyDefinition.defaultValue);
                        onChange.call(t,t.action,true,t.initialObject,t.modifiedObject,serializedValue,t._lastValue,t.scratch);
                    } 
                });

                t._modified = false;
                t.onChangeModified(t._modified);
                t._postModify(true);
            }
        },

        reset: function() {
            this.open(this.initialObject,this.action);
        },

        okToSave: function(noMessages) {
            var t = this;
            var deferred = new Deferred();
            var returnObject = {
                status: false,
                code: ""
            };

            if (t.formNode && !t.formNode.validate()) {
                returnObject.code = 1001; // Invalid field
                deferred.resolve(returnObject);
            } else if (query(".pinkBackground",t.domNode).length > 0) {
                returnObject.code = 1002; // Required field
                deferred.resolve(returnObject);
            } else if (!t.isModified()) {
                returnObject.code = 1003; // No changes
                deferred.resolve(returnObject);
            } else {
                var validOrPromise = true;
                if (typeof t._onValidate === "function") {
                    validOrPromise = t._onValidate.call(t,t.action,t.initialObject,t.modifiedObject,t.scratch);
                }
                if (validOrPromise && t.definition.hasServerOnValidate) {
                    // Execute server-side validation
                    validOrPromise = t._deployPropertyStateChange("ServerOnValidate");
                }

                when(validOrPromise,function(valid) {
                    if (valid) {
                        returnObject.status = true;
                    }
                    deferred.resolve(returnObject);
                    if (!noMessages) {
                        t.publishMessages();
                    } else {
                        t._messages = [];
                    }
                });
            }
            return deferred.promise;
        },

        save: function(skipCheck) {
            var t = this;
            var deferred = new Deferred();
			
            if (!skipCheck) {
                var returnPromise = t.okToSave();
            } else {
                var returnPromise = returnObject = {
                    status: true,
                    code: "",
                    object: null
                };
            }
            when(returnPromise,function(returnObject) {
                if (returnObject.status) {
                    var object = t._stripObject(t.modifiedObject);
                    serverRequest("Editor.Loader",(t.action === "Add" ? "Add" : "Edit"),{
                        editorClass: t.editorClass,
                        object: object
                    }).then(function(status) {
                        returnObject.status = true;
                        returnObject.id = status.id;
                        t._modified = false;
                        t.onSave(); // fire "save" event

                        arrayUtils.forEach(status.messages,function(message) {
                            t.showMessage(message);
                        });
                        t.publishMessages();

                        deferred.resolve(returnObject);
                    },function() {
                        returnObject.status = false;
                        returnObject.code = 2001; // Server error
                        deferred.resolve(returnObject);
                    });
                } else {
                    deferred.resolve(returnObject); // Leave status = false
                }
                t.publishMessages();
            });
            return deferred.promise;
        },

        onSave: function() {},

        saveAndAdd: function() {
            var t = this;
            var deferred = new Deferred();
            if (t.action === "Add") {
                t.save().then(function(returnObject) {
                    if (returnObject.status) {
                        var object = {};
                        if (t.definition) {
                            arrayUtils.forEach((t.definition.propertyDefinitions || []), function(propertyDefinition) {
                                if (propertyDefinition.carryForward) {
                                    object[propertyDefinition.name] = t.modifiedObject[propertyDefinition.name];
                                } else {
                                    object[propertyDefinition.name] = propertyDefinition.defaultValue;
                                }
                            });
                        }
                        t.open(object,"Add",t.getNewObject());
                    }
                    deferred.resolve(returnObject);
                });
            }
            return deferred.promise;
        },

        cancel: function(quitting) {
            // To be called when an action is canceled. Could be used to cancel locks. Return a boolean or a promise that resolves to a boolean. If resolving to false, the cancelation is stopped
			return true;
        },

        deleteEditor: function(id) {
            if (typeof id === "undefined" && this.modifiedObject) {
                id = this.modifiedObject.id;
            }
            return Editor.deleteEditor(this.editorClass,id);
        },

        isModified: function() {
            return this._modified;
        },

        showMessage: function(message) {
            var t = this;
            t._messages.push(message);
        },

        publishMessages: function() {
            var t = this;
            var messageText = "";
            arrayUtils.forEach(t._messages,function(message) {
                messageText += (messageText !== "" ? "<br>" : "")+message;
            });
            if (messageText !== "") {
                topic.publish("editor/message",messageText,t.editorClass);
            }
            t._messages = [];
        },

        getPropertyDefinition: function(propertyName) {
            var propertyDefinition = null;
            arrayUtils.some(this.definition.propertyDefinitions,function(propDefinition) {
                if (propDefinition.name === propertyName) {
                    propertyDefinition = propDefinition;
                    return true;
                }
                return false;
            });
            return propertyDefinition;
        },

        _modifyValue: function(propertyName) {
            var t = this;
            if (t.modifiedObject) {
                var field = t.fields[propertyName];
                var serializedValue = field.serialize();
                if (serializedValue === t.modifiedObject[propertyName]) {
                    t._notify(propertyName);
                    t._postModify(false,propertyName);
                    return;
                }
                if ((typeof field.isValid === "undefined") || field.isValid()) {
                    t._lastValue = t.modifiedObject[propertyName];
                    var validOrPromise = true;
                    var onValidate = t._onValidates[propertyName];
                    if (typeof onValidate === "function") {
                        validOrPromise = onValidate.call(t,t.action,t.initialObject,t.modifiedObject,serializedValue,t._lastValue,t.scratch);
                    }
                    if (validOrPromise && t._hasServerOnValidates[propertyName]) {
                        // Execute server-side validation
                        validOrPromise = t._deployPropertyStateChange("ServerOnValidate"+propertyName,serializedValue);
                    }
                    
                    when(validOrPromise,function(valid) {
                        if (valid) {
                            t.modifiedObject.set(propertyName,serializedValue);

                            var onChange = t._onChanges[propertyName];
                            if (typeof onChange === "function") {
                                onChange.call(t,t.action,false,t.initialObject,t.modifiedObject,serializedValue,t._lastValue,t.scratch);
                            } 
                            if (t._hasServerOnChanges[propertyName]) {
                                t._deployPropertyStateChange("ServerOnChange"+propertyName,serializedValue).then(function() {
                                    t._postModify(false,propertyName);
                                });
                            } else {
                                t._postModify(false,propertyName);
                            }
                        } else {
                            //topic.publish("editor/message",error,t.editorClass);
                            t._notify(propertyName);
                            t._postModify(false,propertyName);
                        }
                    },function(error) {
                        topic.publish("editor/error",error,t.editorClass);
                        t._notify(propertyName);
                    });
                } else {
                    t._postModify(false,propertyName);
                }
            }
        },

        _notify: function(propertyName) {
            var t = this;
            var field = t.fields[propertyName];
            if (field) {
                var serializedValue = t.modifiedObject[propertyName];
                field.deserialize(serializedValue,false);
            }
        },

        _postModify: function(initialize,propertyName) {
            var t = this;
            
            var onChange = t._onChange;
            if (typeof onChange === "function") {
                onChange.call(t,t.action,initialize,t.initialObject,t.modifiedObject,propertyName,t.scratch);
            } 

            arrayUtils.forEach((t.definition.propertyDefinitions || []), function(propertyDefinition) {
                var propName = propertyDefinition.name;
                t.fields[propName].set("disabled",t._computeDisabled(propertyDefinition));
                t.fields[propName].requiredProperty = t._computeRequired(propertyDefinition);
                t.fields[propName].pinkify();
            });

            if (!initialize) {
                t._computeModified();
            } else if (t.onActionDeferred) {
                t.onActionDeferred.resolve();
            }

            t.publishMessages();
        },

        _computeDisabled: function(propertyDefinition) {
            var t = this;
            var propName = propertyDefinition.name;
            var disabled = propertyDefinition.disabled || t.action === "View" || t.disabled;
            if (!disabled) {
                disabled = arrayUtils.some(propertyDefinition.disableActions.toString().split(","),function(action) {
                    return action === t.action;
                });
            }
            if (!disabled && typeof t._isDisableds[propName] === "function") {
                var isDisabled = t._isDisableds[propName];
                disabled = isDisabled.call(t,t.action,t.initialObject,t.modifiedObject);
            }
            return disabled;
        },

        _computeRequired: function(propertyDefinition) {
            var t = this;
            var propName = propertyDefinition.name;
            var required = propertyDefinition.required;
            if (!required && typeof t._isRequireds[propName] === "function") {
                var isRequired = t._isRequireds[propName];
                required = isRequired.call(t,t.action,t.initialObject,t.modifiedObject);
            }
            return required;
        },

        _computeModified: function() {
            var t = this;
            var temp = t._modified;
            t._modified = false;
            arrayUtils.some((t.definition.propertyDefinitions || []),function(propertyDefinition) {
                if (propertyDefinition.trackChanges == 1) {
                    if (propertyDefinition.widget === "thornberry/editor/property/EditableObject" && t.fields[propertyDefinition.name]._editor) {
                        t._modified = t.fields[propertyDefinition.name]._editor._modified;
                    } else if (typeof t.modifiedObject[propertyDefinition.name] === "object") {
                        t._modified = (JSON.stringify(t.modifiedObject.get(propertyDefinition.name)) != JSON.stringify(t.initialObject[propertyDefinition.name]));
                    } else {
                        t._modified = (t.modifiedObject.get(propertyDefinition.name) != t.initialObject[propertyDefinition.name]);
                    }
                }
                return t._modified;
            });
            if (temp !== t._modified) {
                t.onChangeModified(t._modified);
            }
            return t._modified;
        },

        getModified: function() {
            return this._modified;
        },

        onChangeModified: function() {},

        _deployPropertyStateChange: function(method,value) {
            var t = this;
            var strippedObject = t._stripObject(t.modifiedObject);

            topic.publish("editor/preventSave",true);
            return serverRequest("Editor.Loader","OnPropertyStateChange",{
                editorClass: t.editorClass,
                method: method,
                action: t.action,
                initialObject: t.initialObject,
                modifiedObject: strippedObject,
                value: value,
                lastValue: t._lastValue,
                scratch: t.scratch
            }).then(function(result) {
                topic.publish("editor/preventSave",false);

                // Check for changes to modifiedObject
                if (result.modifiedObject) {
                    for (_prop in result.modifiedObject) {
                        if (result.modifiedObject[_prop] != t.modifiedObject[_prop]) {
                            t.modifiedObject.set(_prop,result.modifiedObject[_prop]);
                        }
                    }
                }
                arrayUtils.forEach(result.messages,function(message) {
                    t.showMessage(message);
                });
                t.scratch = result.scratch;
                return result.returnValue;
            },function() {
                topic.publish("editor/preventSave",false);
            });
        },

        _stripObject: function(object) {
            var strippedObject = {};
            for (_prop in object) {
                if (typeof object[_prop] !== "function" && _prop.toString()[0] !== "_") {
                    strippedObject[_prop] = object[_prop];
                }
            }
            return strippedObject;
        },

        callSuperMethod: function(className,methodName,arguments) {
            var t = this;
            var definition = t._getSuperDefinition(className);
            var method = t._getMethod(definition[methodName]);
            return method.apply(t,arguments);
        },

        callSuperPropertyMethod: function(className,propertyName,methodName,arguments) {
            var t = this;
            var returnValue;
            var definition = t._getSuperDefinition(className);
            var propertyDefinition = t.getPropertyDefinition.call({definition:definition},propertyName);
            if (propertyDefinition) {
                var method = t._getMethod(propertyDefinition[methodName]);
                returnValue = method.apply(t,arguments);
            }
            return returnValue;
        },

        _getSuperDefinition: function(className) {
            var t = this;
            var definition = null;
            for (var name in t.definition._super) {
                var pieces = name.split(".");
                if (name === className || pieces[pieces.length-1] === className) {
                    definition = t.definition._super[name];
                    break;
                }
            }
            return definition;
        },

        _getMethod: function(methodText) {
            // TODO: consider initializing the methods once instead of every time they're called (potentially with every field change)
            var t = this;
            var modulePathArray = arrayUtils.map(t.definition.moduleDefinitions,function(definition) {return definition.path});
            var method = null;
            require(modulePathArray,function() {
                // Should evaluate synchronously
                for (var i = 0; i < arguments.length; i++) {
                    eval("var "+t.definition.moduleDefinitions[i].name+" = arguments[i];");
                }

                try {
                    eval("method = "+methodText);
                } catch (e) {
                    console.error(e);
                    method = null;
                }
            });
            return method;
        },
        
        _setDisabledAttr: function(value) {
            var temp = this.disabled;
            this.disabled = value;
            if (temp !== value) {
                this._postModify();
            }
        }
    });

    Editor.deleteEditor = function(editorClass,id) {
        return serverRequest("Editor.Loader","Delete",{
            editorClass: editorClass,
            id: id
        });
    };

    return Editor;
});