define([
	'dojo/_base/declare',
	'dojo/_base/array',
	'dojo/on',
	'dojo/dom-construct',
	'dojo/dom-class',
	'dijit/layout/ContentPane'
], function(declare, array, on, domConstruct, domClass, ContentPane) {
	return declare([ContentPane], {
		titleNode: null,
		
		title: '',
		_setTitleAttr: { node: "titleNode", type: "innerText" },
		
		breadcrumbs: null,
		_setBreadcrumbsAttr: function(breadcrumbs){
			var breadcrumbsNode = this.breadcrumbsNode;
			var breadcrumbHandlers = this.__breadcrumbEventHandlers;
			var registeredOperations = this.registeredOperations;
			this._set('breadcrumbs', breadcrumbs);
			
			while (breadcrumbHandlers.length > 0) {
				breadcrumbHandlers.pop().remove();
			}
			domConstruct.empty(breadcrumbsNode);
			
			array.forEach(breadcrumbs, function(breadcrumb, index) {
				var crumbNode, linkNode;
				if (index !== 0) {
					domConstruct.create('span', {innerHTML: '&nbsp;&gt;&nbsp;'}, breadcrumbsNode);
				}

				var hasAction = false;
				if (breadcrumb.operation && registeredOperations[breadcrumb.operation]) {
					hasAction = true;
				}
				crumbNode = domConstruct.create((hasAction ? 'a' : 'span'), {innerHTML: breadcrumb.text}, breadcrumbsNode);

				if (hasAction) {
					domClass.add(crumbNode, 'breadcrumb-item');
					breadcrumbHandlers.push(on(crumbNode,'click',registeredOperations[breadcrumb.operation]));
				}
			});
		},
		__breadcrumbEventHandlers: null,
		
		constructor: function() {
			this.__breadcrumbEventHandlers = [];
		},
			
		exitLabel: ' Exit',
		_setExitLabelAttr: function(exitLabel) {
			if (typeof exitLabel === 'string') {
				this.exitLabelNode.innerHTML = ' ' + exitLabel;
			}
		},

		showExit: false,
		_setShowExitAttr: function(showExit) {
			if (showExit) {
				domClass.remove(this.exitNode, "hidden");
			} else {
				domClass.add(this.exitNode, "hidden");
			}
		},
		
		buildRendering: function(){
			this.inherited(arguments);
			
			this.breadcrumbsNode = domConstruct.create('div');
			domClass.add(this.breadcrumbsNode,"breadcrumbs");
			this.titleNode = domConstruct.create('div');
			domClass.add(this.titleNode,"workspace-title");
			this.exitNode = domConstruct.create('div', {
				onclick: 'startWindow.updateAccountInHeader(); top.close();',
				style: 'margin-right: 0px;'});
			domClass.add(this.exitNode,"hidden icon");
			domConstruct.create("img", {
				alt: 'Exit',
				src: '/images/MonoIcons/black/sign-out32.png',
				style: 'opacity: 0.6;height:20px;width:20px',
				title: 'Exit',
			}, this.exitNode);
			this.exitLabelNode = domConstruct.create("label", {
				style: 'vertical-align: 20%'
			}, this.exitNode);
		},
		
		postCreate: function(){
			var titleBreadcrumbs = domConstruct.create('div');
			
			domConstruct.place(titleBreadcrumbs, this.containerNode);
			domConstruct.place(this.exitNode, this.containerNode);

			domConstruct.place(this.breadcrumbsNode, titleBreadcrumbs);
			domConstruct.place(this.titleNode, titleBreadcrumbs);
		},
		
		updateContent: function(contentObj) {
			var data = contentObj || {};
			
			if (data.title) {
				this.set('title', data.title);
			}
			if (data.breadcrumbs) {
				this.set('breadcrumbs', data.breadcrumbs);
			}
		},
		
		registeredOperations: {},
		registerOperation: function(name, func) {
			this.registeredOperations[name] = func;
		}
	});
});