define(['dojo/_base/declare', 'dojo/_base/lang', 'dojo/dom-class', 'dojo/dom-style', 'dijit/focus', 'dijit/_WidgetBase', 'dijit/_TemplatedMixin', 'dijit/_Contained', 'dijit/_CssStateMixin', 'dojo/text!./templates/item.html'], function (declare, lang, domClass, domStyle, focus, _WidgetBase, _TemplatedMixin, _Contained, _CssStateMixin, template) {
	return declare([_WidgetBase, _TemplatedMixin, _Contained, _CssStateMixin], {
		// summary:
		//		Single navigation item within a menu.   This class is used internally
		//		by ndoc/nav and should not be accessed directly.
		// tags:
		//		private

		// item: [const] Item
		//		the dojo.data entry this tree represents
		item: null,

		// isNavNode: [protected] Boolean
		//		Indicates that this is an NDocNavItem.   Used by `ndoc/nav/menu` only,
		//		should not be accessed directly.
		isNavNode: true,

		// label: String
		// Text of this navigation item node
		label: "",
		_setLabelAttr: function (val) {
			this.labelNode[this.labelType == "html" ? "innerHTML" : "innerText" in this.labelNode ? "innerText" : "textContent"] = val;
			this._set("label", val);
		},

		// labelType: [const] String
		//		Specifies how to interpret the label.  Can be "html" or "text".
		labelType: "text",

		// isExpandable: [private] Boolean
		//		This node has children, so show the expando node (+ sign)
		isExpandable: null,

	   // Tooltip is defined in _WidgetBase but we need to handle the mapping to DOM here
		_setTooltipAttr: {
			node: "rowNode",
			type: "attribute",
			attribute: "title"
		},

		// For hover effect for tree node, and focus effect for label
		cssStateNodes: {
			rowNode: IS_MOBILE ? "ndocNavRowMobile": "ndocNavRow"
		},

		baseClass: "ndocNavItem",

		templateString: IS_MOBILE ? template.replace("ndocNavRow", "ndocNavRowMobile") : template,
		buildRendering: function () {
			this.inherited(arguments);
			// set icon and label class and style based on item
			this._updateItemClasses(this.item);
		},

		_updateItemClasses: function (item) {
			// summary:
			//		Set appropriate CSS classes for icon and label dom node
			//		(used to allow for item updates to change respective CSS)
			// tags:
			//		private
			var nav = this.nav;
			this._applyClassAndStyle(item, "expando", "Expando");
			this._applyClassAndStyle(item, "icon", "Icon");
			this._applyClassAndStyle(item, "favorite", "Favorite");
			this._applyClassAndStyle(item, "label", "Label");
			this._applyClassAndStyle(item, "row", "Row");

			nav._startPaint(true);		// signifies paint started and finished (synchronously)
		},

		_applyClassAndStyle: function (item, lower, upper) {
			// summary:
			//		Set the appropriate CSS classes and styles for labels, icons and rows.
			//
			// item:
			//		The data item.
			//
			// lower:
			//		The lower case attribute to use, e.g. 'icon', 'label' or 'row'.
			//
			// upper:
			//		The upper case attribute to use, e.g. 'Icon', 'Label' or 'Row'.
			//
			// tags:
			//		private

			var clsName = "_" + lower + "Class";
			var nodeName = lower + "Node";
			var oldCls = this[clsName];

			this[clsName] = this.nav["get" + upper + "Class"](item);
			domClass.replace(this[nodeName], this[clsName] || "", oldCls || "");

			domStyle.set(this[nodeName], this.nav["get" + upper + "Style"](item) || {});
		},
		
		setSelected: function(/*Boolean*/ selected){
			// summary:
			//		A Tree has a (single) currently selected node.
			//		Mark that this node is/isn't that currently selected node.
			// description:
			//		In particular, setting a node as selected involves setting tabIndex
			//		so that when user tabs to the tree, focus will go to that node (only).
			domClass.toggle(this.rowNode, "ndocNavRowSelected", selected);
		},

		focus: function(){
			focus.focus(this.focusNode);
		}
	});
});