define([
	'dojo/_base/declare',
	'dojo/_base/lang',
	'dojo/_base/array',
	'dojo/aspect',
	'dojo/dom',
	'dojo/dom-class',
	'dojo/dom-style',
	'dojo/Deferred',
	'dojo/fx',
	'dojo/keys',
	'dojo/on',
	'dojo/promise/all',
	'dojo/string',
	'dojo/topic',
	'dojo/touch',
	'dojo/when',
	'dijit/a11yclick',
	'dijit/form/TextBox',
	'dijit/registry',
	'dijit/_WidgetBase',
	'dijit/_TemplatedMixin',
	'dijit/_KeyNavMixin',
	'dijit/_Container',
	'dojo/text!./templates/menu.html',
	'./item',
	'dojo/query!css2'
], function (declare, lang, array, aspect, dom, domClass, domStyle, Deferred, fxUtils, keys, on, all, string, topic, touch, when, a11yclick, TextBox, registry, _WidgetBase, _TemplatedMixin, _KeyNavMixin, _Container, template, NDocNavItem) {
	return declare([_WidgetBase, _TemplatedMixin, _Container, _KeyNavMixin], {
		// summary:
		//		This widget displays hierarchical data from a store as a sliding drill-down.

		baseClass: "ndocNav",

		// model: [const] dijit/tree/model
		//		Interface to read tree data, get notifications of changes to tree data,
		//		and for handling drop operations (i.e drag and drop onto the tree)
		model: null,

		// state: [private] String
		//		Dynamic loading-related stuff.
		//		When an empty folder node appears, it is "NotLoaded" first,
		//		then after dojo.data query it becomes "Loading" and, finally "Loaded"
		state: "NotLoaded",

		templateString: template,

		_publish: function(/*String*/ topicName, /*Object*/ message){
			// summary:
			//		Publish a message for this widget/topic
			topic.publish(this.id, lang.mixin({nav: this, event: topicName}, message || {}));	// publish
		},

		postMixInProperties: function(){
			this.inherited(arguments);
			
			this.nav = this;

			this._itemNodesMap = {};

			// Deferred that resolves when all the children have loaded.
			this.expandNodeDeferred = new Deferred();

			// Promise that resolves when all pending operations complete.
			this.pendingCommandsPromise = this.expandNodeDeferred.promise;
		},

		postCreate: function(){
			this.inherited(arguments);

			this.searchBox = new TextBox({
				name: 'navSearch',
				style: 'width: 100%',
				value: "",
				placeHolder: "Find a function"
			}).placeAt(this.searchNode);
			
			// Catch events on NDocNavNodes
			var self = this;
			this.own(
				on(this.searchBox, "input", lang.hitch(this, "_onSearchInput")),
				on(this.searchBox, "mouseup", lang.hitch(this, function(evt){
					var _this = this,
						oldValue = this.searchBox.get('value');
					// Wait a bit and check value again to see if value was cleared
					if (oldValue != "") {
						setTimeout(function(){
							var newValue = _this.searchBox.get('value');
							if (newValue == "") {
								_this._onSearchInput(evt);
							}
						}, 1);
					}
				})),

				on(this.containerNode, on.selector(IS_MOBILE ? ".ndocNavRowMobile": ".ndocNavRow", a11yclick.press), function(evt){
					self._onNodePress(registry.getEnclosingWidget(this), evt);
				}),
				on(this.containerNode, on.selector(IS_MOBILE ? ".ndocNavRowMobile": ".ndocNavRow", a11yclick), function(evt){
					self._onClick(registry.getEnclosingWidget(this), evt);
				}),
				on(this.backButtonNode, a11yclick, function(evt){
					self._onBack(evt);
				})
			);

			this.own(
				aspect.after(this.model, "onFavoriteChange", lang.hitch(this, "_onFavoriteChange"), true)
			);

			this._load();

			// onLoadDeferred should fire when all commands that are part of initialization have completed.
			// It will include all the set("paths", ...) commands that happen during initialization.
			this.onLoadDeferred = this.pendingCommandsPromise;

			this.onLoadDeferred.then(lang.hitch(this, "onLoad"));
		},

		onLoad: function(){
			// summary:
			//		Called when menu finishes loading and expanding.
			// description:
			//		If persist == true the loading may encompass many levels of fetches
			//		from the data store, each asynchronous.   Waits for all to finish.
			// tags:
			//		callback
		},

		_load: function(){
			// summary:
			//		Initial load of the navigation menu.
			//		Load root node (possibly hidden) and it's children.
			this.model.getRoot(
				lang.hitch(this, function(item){
					var rn = (this.rootNode = this._createNavItem({
						item: item,
						nav: this,
						label: this.label || this.getLabel(item),
						labelType: this.model.labelType || "text"
					}));
					// make root node invisible
					rn.rowNode.style.display = "none";
					rn.labelNode.setAttribute("role", "presentation");
						
					var identity = this.model.getIdentity(item);
					if(this._itemNodesMap[identity]){
						this._itemNodesMap[identity].push(rn);
					}else{
						this._itemNodesMap[identity] = [rn];
					}

					// Load top level children
					this._expandNode(rn).then(lang.hitch(this, function(){
						// Then, select the nodes specified by params.paths[].
						this.expandNodeDeferred.resolve(true);
					}));
				}),
				lang.hitch(this, function(err){
					console.error(this, ": error loading root: ", err);
				})
			);
		},

		// searchDelay: Integer
		//		Delay in milliseconds between when user types something and we start
		//		searching based on that value
		searchDelay: 200,

		// searchAttr: String
		//		Search for items in the data store where this attribute (in the item)
		//		matches what the user typed
		searchAttr: "fullName",

		// queryExpr: String
		//		This specifies what query is sent to the data store,
		//		based on what the user has typed.  Changing this expression will modify
		//		whether the results are only exact matches, a "starting with" match,
		//		etc.
		//		`${0}` will be substituted for the user text.
		//		`*` is used for wildcards.
		//		`${0}*` means "starts with", `*${0}*` means "contains", `${0}` means "is"
		queryExpr: "*${0}*",

		_onSearchInput: function(/*Event*/ evt){
			// summary:
			//		Start a new search, cancel the running one if needed.
			var key = evt.charOrCode;

			var doSearch = false;
			this._prev_key_backspace = false;

			switch(key){
				case keys.DELETE:
				case keys.BACKSPACE:
					this._prev_key_backspace = true;
					this._maskValidSubsetError = true;
					doSearch = true;
					break;

				default:
					// Non char keys (F1-F12 etc..) shouldn't start a search..
					// Ascii characters and IME input (Chinese, Japanese etc.) should.
					//IME input produces keycode == 229.
					doSearch = typeof key == 'string' || key == 229 || typeof key == 'undefined';
			}
			if(doSearch){
				// need to wait a tad before start search so that the event
				// bubbles through DOM and we have value visible
				this.searchTimer = this.defer("_startSearch", 1);
			}
		},

		_startSearch: function(){
			// summary:
			//		Starts a search for elements matching text (text=="" means to return all items),
			//		and calls onSearch(...) when the search completes, to display the results.

			this._abortQuery();
			var text = this.searchBox.focusNode.value;
			if (text == "") {
				this._expandNode(this.rootNode, true);
				return;
			}
			var
				_this = this,
				// Setup parameters to be passed to store.query().
				// Create a new query to prevent accidentally querying for a hidden
				// value from FilteringSelect's keyField
				query = {},
				options = {
					queryOptions: {		// remove for 2.0
						ignoreCase: true,
						deep: true
					}
				},
				qs = string.substitute(this.queryExpr, [text.replace(/([\\\*\?])/g, "\\$1")]),
				q,
				startQuery = function(){
					var resPromise = _this._fetchHandle = _this.model.permissionStore.query(query, options);
					if(q !== _this._lastQuery){
						return;
					} // avoid getting unwanted notify
					when(resPromise, function(res){
						_this._fetchHandle = null;
						if(q === _this._lastQuery){ // avoid getting unwanted notify
							when(resPromise.total, function(total){
								res.total = total;
								_this.onSearch(res, query, options);
							});
						}
					}, function(err){
						_this._fetchHandle = null;
						if(!_this._cancelingQuery){	// don't treat canceled query as an error
							console.error(_this.declaredClass + ' ' + err.toString());
						}
					});
				};

			// Generate query
			// Query on searchAttr is a regex for benefit of dojo/store/Memory,
			// but with a toString() method to help dojo/store/JsonRest.
			// Search string like "Co*" converted to regex like /^Co.*$/i.
			q = this._patternToRegExp(qs);
			q.toString = function(){ return qs; };

			// set _lastQuery, *then* start the timeout
			// otherwise, if the user types and the last query returns before the timeout,
			// _lastQuery won't be set and their input gets rewritten
			this._lastQuery = query[this.searchAttr] = q;
			this._queryDeferHandle = this.defer(startQuery, this.searchDelay);
		},

		_abortQuery: function(){
			// stop in-progress query
			if(this.searchTimer){
				this.searchTimer = this.searchTimer.remove();
			}
			if(this._queryDeferHandle){
				this._queryDeferHandle = this._queryDeferHandle.remove();
			}
			if(this._fetchHandle){
				if(this._fetchHandle.abort){
					this._cancelingQuery = true;
					this._fetchHandle.abort();
					this._cancelingQuery = false;
				}
				if(this._fetchHandle.cancel){
					this._cancelingQuery = true;
					this._fetchHandle.cancel();
					this._cancelingQuery = false;
				}
				this._fetchHandle = null;
			}
		},

		onSearch: function(results, query, options){
			// summary:
			//		Callback when a search completes.
			//
			// results: Object
			//		An array of items from the originating _SearchMixin's store.
			//
			// query: Object
			//		A copy of the originating _SearchMixin's query property.
			//
			// options: Object
			//		The additional parameters sent to the originating _SearchMixin's store, including: start, count, queryOptions.
			//
			// tags:
			//		callback
			this._displaySearchResults(results, true);
		},

		_displaySearchResults: function(results, /*Boolean*/ dontFocus){
			// summary:
			//		Called when the user has requested to expand the node
			// returns:
			//		Promise that resolves when the node is loaded and opened

			var model = this.model,
				_this = this;

			_this.markProcessing();
			this.collapse();

			// Setup deferred to signal when the load and expand are finished.
			// Save that deferred in this._expandDeferred as a flag that operation is in progress.
			results._loadDeferred = new Deferred();

			// Display the children
			_this.setChildItems(results, dontFocus, 1).then(function(){
				results._loadDeferred.resolve(results);
			});

			// Expand the node after data has loaded
			var def = results._loadDeferred.then(lang.hitch(this, function(){
				var node = {
					item: {
						id: 'HOME',
						name: 'Search Results'
					}
				};
				var def2 = this.expand();

				// seems like these should delayed until expand() completes
				this._state(node, false);

				return def2;
			}));

			this._startPaint(def);	// after this finishes, need to reset widths of TreeNodes

			return def;	// dojo/promise/Promise
		},

		_expandNode: function(/*NDocNavItem*/ node, /*Boolean*/ dontFocus){
			// summary:
			//		Called when the user has requested to expand the node
			// returns:
			//		Promise that resolves when the node is loaded and opened

			var model = this.model,
				item = node.item,
				_this = this;

			_this.markProcessing();
			this.collapse();

			// Setup deferred to signal when the load and expand are finished.
			// Save that deferred in this._expandDeferred as a flag that operation is in progress.
			node._loadDeferred = new Deferred();

			// Get the children
			model.getChildren(
				item,
				function(items){
					_this.unmarkProcessing();

					// Display the children
					_this.setChildItems(items, dontFocus).then(function(){
						node._loadDeferred.resolve(items);
					});
				},
				function(err){
					console.error(_this, ": error loading " + node.label + " children: ", err);
					node._loadDeferred.reject(err);
				}
			);

			// Expand the node after data has loaded
			var def = node._loadDeferred.then(lang.hitch(this, function(){
				var def2 = this.expand();

				// seems like these should delayed until node.expand() completes, but left here for back-compat about
				// when this.isOpen flag gets set (ie, at the beginning of the animation)
				this.onOpen(item, node);
				this._state(node, true);

				return def2;
			}));

			this._startPaint(def);	// after this finishes, need to reset widths of TreeNodes

			return def;	// dojo/promise/Promise
		},

		setChildItems: function(/* Object[] */ items, /*Boolean*/ dontFocus, /*Boolean*/ isSearchResults){
			// summary:
			//		Sets the child items of this menu, removing/adding nodes
			//		from current children to match specified items[] array.
			// returns:
			//		Promise that resolves after all previously opened children
			//		have been expanded again (or fires instantly if there are no such children).

			var nav = this,
				model = nav.model,
				firstChild = null,
				defs = [];	// list of deferreds that need to fire before I am complete


			// Orphan all my existing children.
			// If items contains some of the same items as before then we will reattach them.
			// Don't call this.removeChild() because that will collapse the tree etc.
			var oldChildren = this.getChildren();
			array.forEach(oldChildren, function(child){
				_Container.prototype.removeChild.call(this, child);
			}, this);

			this.state = "Loaded";

			if(items && items.length > 0){
				// Create NDocNavItem widget for each specified menu node, unless one already
				// exists and isn't being used
				array.forEach(items, function(item){    // MARKER: REUSE NODE
					var id = model.getIdentity(item),
						existingNodes = nav._itemNodesMap[id],
						node;
					if(!isSearchResults && existingNodes){
						for(var i = 0; i < existingNodes.length; i++){
							if(existingNodes[i] && !existingNodes[i].getParent()){
								node = existingNodes[i];
								break;
							}
						}
					}
					if(!node){
						node = nav._createNavItem({
							item: item,
							nav: nav,
							isExpandable: model.getExpandable(item),
							isFavoriteable: model.getFavoriteable(item),
							label: isSearchResults ? item.fullName : nav.getLabel(item),
							labelType: (model && model.labelType) || "text",
							tooltip: nav.getTooltip(item),
							ownerDocument: nav.ownerDocument,
							lang: nav.lang
						});
						if (!isSearchResults) {
							if(existingNodes){
								existingNodes.push(node);
							}else{
								nav._itemNodesMap[id] = [node];
							}
						}
					}
					if (!firstChild) { firstChild = node; }
					this.addChild(node);
				}, this);
			} else {
				var empty = nav._createNavItem({
							item: null,
							nav: nav,
							isExpandable: false,
							isFavoriteable: false,
							label: '(No functions found)',
							labelType: "text",
							ownerDocument: nav.ownerDocument,
							lang: nav.lang
						});
				this.addChild(empty);
				if (!firstChild) { firstChild = empty; }
			}

			if(!dontFocus && firstChild && firstChild.isNavNode){
				this.focusNode(firstChild);
			}
				
			var def = all(defs);
			nav._startPaint(def);		// to reset NDocNavItem widths after an item is added/removed from the NavMenu
			return def;		// dojo/promise/Promise
		},

		expand: function(){
			// summary:
			//		Show my children
			// returns:
			//		Promise that resolves when expansion is complete

			// If there's already an expand in progress or we are already expanded, just return
			if(this._expandDeferred){
				return this._expandDeferred;		// dojo/promise/Promise
			}
			// cancel in progress collapse operation
			if(this._collapseDeferred){
				this._collapseDeferred.cancel();
				delete this._collapseDeferred;
			}
			
			// Animations aren't working currently; Just create and resolve a deferred

			// Deferred that fires when expand is complete
			var def = (this._expandDeferred = new Deferred(function(){}));

			def.resolve(true);

			return def;		// dojo/promise/Promise
		},
		
		collapse: function(){
			// summary:
			//		Collapse this node (if it's expanded)
			// returns:
			//		Promise that resolves when collapse is complete

			if(this._collapseDeferred){
				// Node is already collapsed, or there's a collapse in progress, just return that Deferred
				return this._collapseDeferred;
			}

			// cancel in progress expand operation
			if(this._expandDeferred){
				this._expandDeferred.cancel();
				delete this._expandDeferred;
			}

			// Animations aren't working currently; Just create and resolve a deferred

			// Deferred that fires when expand is complete
			var def = (this._collapseDeferred = new Deferred(function(){}));

			def.resolve(true);

			return def;		// dojo/promise/Promise
		},

		///////////////////////////////////////////////////////
		// Functions for converting an item to a NDocNavItem
		getLabel: function ( /*dojo/data/Item*/ item) {
			// summary:
			//		Overridable function to get the label for a tree node (given the item)
			// tags:
			//		extension
			return this.model.getLabel(item); // String
		},

		getIconClass: function ( /*dojo/data/Item*/ item, /* Boolean */ forHeader) {
			// summary:
			//		Overridable function to return CSS class name to display icon
			// tags:
			//		extension
			if (forHeader) {
				return item.id.split(':')[0];
			}
			return (!!item && (array.indexOf(item.parents, 'START') != -1)) ? (item.id == 'Analytics' ? 'HOME': item.id) : ''; // String
		},

		getExpandoClass: function ( /*dojo/data/Item*/ item) {
			// summary:
			//		Overridable function to return CSS class name to display icon
			// tags:
			//		extension
			return this.model.getExpandable(item) ? 'ndocNavItemExpandIcon' : ''; // String
		},

		getFavoriteClass: function ( /*dojo/data/Item*/ item) {
			// summary:
			//		Overridable function to return CSS class name to display favorite icon
			// tags:
			//		extension
			return this.model.getFavoriteable(item) ? (this.model.getFavorited(item) ? 'ndocNavItemFavorited' : 'ndocNavItemFavoriteEmpty') : ''; // String
		},

		getLabelClass: function(/*===== item, opened =====*/){
			// summary:
			//		Overridable function to return CSS class name to display label
			// item: dojo/data/Item
			// opened: Boolean
			// returns: String
			//		CSS class name
			// tags:
			//		extension
		},

		getRowClass: function(/*===== item, opened =====*/){
			// summary:
			//		Overridable function to return CSS class name to display row
			// item: dojo/data/Item
			// opened: Boolean
			// returns: String
			//		CSS class name
			// tags:
			//		extension
		},

		getFavoriteStyle: function(/*===== item, opened =====*/){
			// summary:
			//		Overridable function to return CSS styles to display favorite icon
			// item: dojo/data/Item
			// opened: Boolean
			// returns: Object
			//		Object suitable for input to dojo.style() like {backgroundImage: "url(...)"}
			// tags:
			//		extension
		},

		getIconStyle: function(/*===== item, opened =====*/){
			// summary:
			//		Overridable function to return CSS styles to display icon
			// item: dojo/data/Item
			// opened: Boolean
			// returns: Object
			//		Object suitable for input to dojo.style() like {backgroundImage: "url(...)"}
			// tags:
			//		extension
		},

		getExpandoStyle: function(/*===== item, opened =====*/){
			// summary:
			//		Overridable function to return CSS styles to display icon
			// item: dojo/data/Item
			// opened: Boolean
			// returns: Object
			//		Object suitable for input to dojo.style() like {backgroundImage: "url(...)"}
			// tags:
			//		extension
		},

		getLabelStyle: function(/*===== item, opened =====*/){
			// summary:
			//		Overridable function to return CSS styles to display label
			// item: dojo/data/Item
			// opened: Boolean
			// returns:
			//		Object suitable for input to dojo.style() like {color: "red", background: "green"}
			// tags:
			//		extension
		},

		getRowStyle: function(/*===== item, opened =====*/){
			// summary:
			//		Overridable function to return CSS styles to display row
			// item: dojo/data/Item
			// opened: Boolean
			// returns:
			//		Object suitable for input to dojo.style() like {background-color: "#bbb"}
			// tags:
			//		extension
		},

		getTooltip: function(/*dojo/data/Item*/ /*===== item =====*/){
			// summary:
			//		Overridable function to get the tooltip for a tree node (given the item)
			// tags:
			//		extension
			return "";	// String
		},

		getItemModules: function(/*dojo/data/Item*/ item){
			// summary:
			//		Overridable function to get the tooltip for a tree node (given the item)
			// tags:
			//		extension
			var _item = this._openItem || item,
				_queue = [],
				modules = [];
			while(_item.parents || _queue.length) {
				_queue = _queue.concat(_item.parents);
				if (_item.intent == 'launch-module') {
					modules.push(_item.id);
				}
				_item = this.model.permissionStore.get(_queue.splice(0,1)) || {};
			}
			return modules;	// String
		},

		/////////// Keyboard and Mouse handlers ////////////////////

		_onDownArrow: function(/*Event*/ evt, /*NDocNavItem*/ node){
			// summary:
			//		down arrow pressed; get next visible node, set focus there
			if (node && node.isNavNode) {
				var nextNode = this._getNext(node);
				if(nextNode && nextNode.isNavNode){
					this.focusNode(nextNode);
				}
			} else {
				this.focusNode();
			}
		},

		_onUpArrow: function(/*Event*/ evt, /*NDocNavItem*/ node){
			// summary:
			//		Up arrow pressed; move to previous visible node

			// if younger siblings
			if (node && node.isNavNode) {
				var previousSibling = node.getPreviousSibling();
				if(previousSibling){
					node = previousSibling;
				}
			}
			
			if(node && node.isNavNode){
				this.focusNode(node);
			}
		},
		
		_onRightArrow: function(/*Event*/ evt, /*NDocNavItem*/ node){
			// summary:
			//		Right arrow pressed; go to child node

			// don't do anything if this is while their typing their search query
			if (this.searchBox.focusNode != evt.srcElement) {
				// if not expanded, expand, else move to 1st child
				if(node && node.isNavNode && node.isExpandable){
					this._expandNode(node);
				}
			}
		},

		_onLeftArrow: function(/*Event*/ evt, /*NDocNavItem*/ node){
			// summary:
			//		Left arrow pressed.
			//		Move to parent.
			
			// don't do anything if this is while their typing their search query
			if (this.searchBox.focusNode != evt.srcElement) {
				if(node && node.isNavNode){
					this._onBack(evt);
				}
			}
		},

		focusLastChild: function(){
			// summary:
			//		End key pressed; go to last visible node.

			var node = this._getLast();
			if(node && node.isNavNode){
				this.focusNode(node);
			}
		},

		_getFirst: function(){
			// summary:
			//		Returns the first child.
			// tags:
			//		abstract extension
			var node = this.rootNode;
			var c = this.getChildren();
			if(c.length){
				node = c[0];
			}
			return node;
		},

		_getLast: function(){
			// summary:
			//		Returns the last descendant.
			// tags:
			//		abstract extension
			var node = this.rootNode;
			var c = this.getChildren();
			if(c.length){
				node = c[c.length - 1];
			}
			return node;
		},


		// Tree only searches forward so dir parameter is unused
		_getNext: function(node){
			// summary:
			//		Returns the next descendant, compared to "child".
			// node: Widget
			//		The current widget
			// tags:
			//		abstract extension

			return node.getNextSibling();
		},

		// Implement _KeyNavContainer.childSelector, to identify which nodes are navigable
		childSelector: ".ndocNavRow",

		isExpandoNode: function(node, widget){
			// summary:
			//		check whether a dom node is the expandoNode for a particular TreeNode widget
			return dom.isDescendant(node, widget.expandoNode) || dom.isDescendant(node, widget.expandoNodeText);
		},

		isFavoriteNode: function(node, widget){
			// summary:
			//		check whether a dom node is the expandoNode for a particular TreeNode widget
			return dom.isDescendant(node, widget.iconNode);
		},

		_onNodePress: function(/*NDocNavItem*/ nodeWidget, /*Event*/ e){
			// Touching a node should focus it, even if you touch the expando node or the edges rather than the label.
			this.focusNode(nodeWidget);
		},

		__click: function(/*NDocNavItem*/ nodeWidget, /*Event*/ e, /*Boolean*/doOpen, /*String*/func){
			var domElement = e.target,
				isExpandoClick = this.isExpandoNode(domElement, nodeWidget),
				isFavoriteClick = this.isFavoriteNode(domElement, nodeWidget);

			if(nodeWidget.isFavoriteable && isFavoriteClick){
				// favorite node was clicked
				this._onFavoriteClick({node: nodeWidget});
			}else if(nodeWidget.isExpandable && isExpandoClick){
				// expando node was clicked, or label of a folder node was clicked; open it
				this._onExpandoClick({node: nodeWidget});
			}else if(nodeWidget.item.intent){
				var item = nodeWidget.item,
					intent = item.intent;
				if (intent == "launch-module") {
					window.OpenModule(item.id);
				} else if (intent == "launch-tab") {
					if (item.id == "Analytics" || item.id.substring(0,9) == "Dashboard") {
						window.createAnalyticsWorkspace(item.name);
					} else {
						console.log("unhandled intent on item("+item.id+"): "+intent);
					}
				}else if (intent == "launch-function") {
					window.OpenRoutine(this.getItemModules(item), item.id);
				} else {
					console.log("unhandled intent: "+intent);
				}
				this._expandNode(this.rootNode, true);
				hideMenu();
			}else if(nodeWidget.isExpandable){
				// expando node was clicked, or label of a folder node was clicked; open it
				this._onExpandoClick({node: nodeWidget});
			}else{
				this._publish("execute", { item: nodeWidget.item, node: nodeWidget, evt: e });
				this[func](nodeWidget.item, nodeWidget, e);
				this.focusNode(nodeWidget);
			}
			e.stopPropagation();
			e.preventDefault();
		},
		_onClick: function(/*NDocNavItem*/ nodeWidget, /*Event*/ e){
			// summary:
			//		Translates click events into commands for the controller to process
			this.__click(nodeWidget, e, false, 'onClick');
		},

		_onExpandoClick: function(/*Object*/ message){
			// summary:
			//		User clicked the > icon; expand my children.
			var node = message.node;

			this._expandNode(node);
		},

		_onFavoriteClick: function(/*Object*/ message){
			// summary:
			//		User clicked the > icon; expand my children.
			var node = message.node,
				isFavorited = this.model.getFavorited(node.item);
				
			this.model.setFavorited(node.item, !isFavorited);

			node._updateItemClasses(node.item);
		},

		_onBack: function(/*Event*/ e){
			// summary:
			//		Go back
			var parentId = this._openItem && this._openItem.parents && this._openItem.parents[0] || 'START',
				existingNodes = this._itemNodesMap[parentId],
				node;
			if(existingNodes){
				for(var i = 0; i < existingNodes.length; i++){
					if(existingNodes[i] && !existingNodes[i].getParent()){
						node = existingNodes[i];
						break;
					}
				}
			}
			if(!node){
				var item = this.model.permissionStore.query({id: parentId})[0];
				node = this._createNavItem({
					item: item,
					nav: this,
					isExpandable: this.model.getExpandable(item),
					label: this.getLabel(item),
					labelType: (this.model && this.model.labelType) || "text",
					tooltip: this.getTooltip(item),
					ownerDocument: this.ownerDocument,
					lang: this.lang
				});
				if(existingNodes){
					existingNodes.push(node);
				}else{
					this._itemNodesMap[parentId] = [node];
				}
			}
		   this.searchBox.set('value', '');
		   this._expandNode(node);
		},

		onClick: function(/*===== item, node, evt =====*/){
			// summary:
			//		Callback when a tree node is clicked
			// item: Object
			//		Object from the dojo/store corresponding to this TreeNode
			// node: TreeNode
			//		The TreeNode itself
			// evt: Event
			//		The event
			// tags:
			//		callback
		},
		onOpen: function(/*===== item, node =====*/){
			// summary:
			//		Callback when a node is opened
			// item: dojo/data/Item
			// node: TreeNode
			// tags:
			//		callback
		},
		
		/* Events from the model */
		_onFavoriteChange: function(/*Item*/ item){
			// summary:
			//		Processes notification of a change to an item's scalar values like label
			var model = this.model,
				identity = model.getIdentity(item),
				nodes = this._itemNodesMap[identity];

			if(nodes){
				array.forEach(nodes, function(node){
					node._updateItemClasses(item);
				});
			}

		},

		/////////////// Miscellaneous funcs

		focusNode: function(/* _tree.Node */ node){
			// summary:
			//		Focus on the specified node (which must be visible)
			// tags:
			//		protected

			var scrollLeft = this.domNode.scrollLeft;
			this.focusChild(node);
			this.domNode.scrollLeft = scrollLeft;
		},

		_createNavItem: function ( /*Object*/ args) {
			// summary:
			//		creates a ndoc/nav/item
			// description:
			//		Developers can override this method to define their own NavItem class;
			//		However it will probably be removed in a future release in favor of a way
			//		of just specifying a widget for the label, rather than one that contains
			//		the children too.
			return new NDocNavItem(args);
		},
		
		_state: function(node, expanded){
			// summary:
			//		Query or set expanded state for an node
			if(arguments.length === 1){
				return this._openItem == node.item;
			}else{
				if(expanded){
					this._openItem = node.item;
				}else{
					this._openItem = null;
				}
				this._updateHeader(node);
			}
		},
		
		_updateHeader: function(selectedNode){
			this.selectedNode = selectedNode;
			this.toggleHeader(selectedNode != this.rootNode);
			this.headerTextNode[
				selectedNode.labelType == 'html' ? 
					'innerHTML' : 
					'innerText' in this.headerTextNode ? 
						'innerText' : 
						'textContent'
			] = this.getLabel(selectedNode.item);
			domClass.replace(this.headerNode, 'ndocNavHeader ' + this.getIconClass(selectedNode.item,1));
		},

		toggleHeader: function(toggleOn){
			this.headerNode.style.display = toggleOn ? 'block' : 'none';
		},

		getPath: function(node){
			var path = [];
			while(node && node !== this.rootNode){
				path.unshift(node.item);
				node = node.getParent();
			}
			path.unshift(this.rootNode.item);

			return path;
		},

		markProcessing: function(){
			// summary:
			//		Visually denote that menu is loading data, etc.
			// tags:
			//		private
			this.state = "Loading";
		},

		unmarkProcessing: function(){
			// summary:
			//		Clear markup from markProcessing() call
			// tags:
			//		private
		},

		_patternToRegExp: function(pattern){
			// summary:
			//		Helper function to convert a simple pattern to a regular expression for matching.
			// description:
			//		Returns a regular expression object that conforms to the defined conversion rules.
			//		For example:
			//
			//		- ca*   -> /^ca.*$/
			//		- *ca*  -> /^.*ca.*$/
			//		- *c\*a*  -> /^.*c\*a.*$/
			//		- *c\*a?*  -> /^.*c\*a..*$/
			//
			//		and so on.
			// pattern: string
			//		A simple matching pattern to convert that follows basic rules:
			//
			//		- * Means match anything, so ca* means match anything starting with ca
			//		- ? Means match single character.  So, b?b will match to bob and bab, and so on.
			//		- \ is an escape character.  So for example, \* means do not treat * as a match, but literal character *.
			//
			//		To use a \ as a character in the string, it must be escaped.  So in the pattern it should be
			//		represented by \\ to be treated as an ordinary \ character instead of an escape.

			return new RegExp("^" + pattern.replace(/(\\.)|(\*)|(\?)|\W/g, function(str, literal, star, question){
				return star ? ".*" : question ? "." : literal ? literal : "\\" + str;
			}) + "$", "mi");
		},

		_outstandingPaintOperations: 0,
		_startPaint: function(/*Promise|Boolean*/ p){
			// summary:
			//		Called at the start of an operation that will change what's displayed.
			// p:
			//		Promise that tells when the operation will complete.  Alternately, if it's just a Boolean, it signifies
			//		that the operation was synchronous, and already completed.

			this._outstandingPaintOperations++;
			if(this._adjustWidthsTimer){
				this._adjustWidthsTimer.remove();
				delete this._adjustWidthsTimer;
			}

			var oc = lang.hitch(this, function(){
				this._outstandingPaintOperations--;

				if(this._outstandingPaintOperations <= 0 && !this._adjustWidthsTimer && this._started){
					// Use defer() to avoid a width adjustment when another operation will immediately follow,
					// such as a sequence of opening a node, then it's children, then it's grandchildren, etc.
					this._adjustWidthsTimer = this.defer("_adjustWidths");
				}
			});
			when(p, oc, oc);
		},

		_adjustWidths: function(){
			// summary:
			//		Size container to match widest TreeNode, so that highlighting with scrolling works (#13141, #16132)

			if(this._adjustWidthsTimer){
				this._adjustWidthsTimer.remove();
				delete this._adjustWidthsTimer;
			}

			this.containerNode.style.width = "auto";
			this.containerNode.style.width = this.domNode.scrollWidth > this.domNode.offsetWidth ? "auto" : "100%";
		},

		focus: function(){
			// summary:
			//		Default focus() implementation: focus the previously focused child, or first child.
			//		Some applications may want to change this method to focus the [first] selected child.

			if(this.lastFocusedChild){
				this.focusNode(this.lastFocusedChild);
			}else{
				this.focusFirstChild();
			}
		},
		
		fxDuration: 200
   });
});