define([
    "dojo/_base/declare",
    "dojo/Deferred",
    "dojo/promise/all",
    "thornberry/layout/Controller",
    "thornberry/report/GridView",
    "thornberry/object/ObjectEditor",
    "thornberry/object/template",
    "thornberry/ReportManagerII",
    "thornberry/layout/View",
    "thornberry/object/ObjectModel",
    "dijit/registry",
    "dijit/form/DropDownButton",
    "dijit/DropDownMenu",
    "dijit/form/Button",
    "dijit/MenuItem",
    "dojo/_base/array",
    "dojo/_base/lang",
    "dojo/dom",
    "dojo/dom-style",
    "dojo/store/Memory",
    "thornberry/layout/DragAndDropPane",
    "dojo/dom-construct",
    "thornberry/serverRequest",
    "dojo/on",
    "dojo/aspect",
    "dojo/Stateful",
    "dojo/query",
    "dijit/Tooltip",
    'dojo/text!/help/NDocReport.html',
], function (declare, Deferred, all, Controller, GridView, ObjectEditor, template, ReportManager, View, ObjectModel, registry, DropDownButton, DropDownMenu, Button, MenuItem, arrayUtils, lang, dom, domStyle, Memory, DragAndDropPane, domConstruct, serverRequest, on, aspect, Stateful, query, Tooltip, helpTemplate) {
    // Define global function openFile() for formatter use
	openFile = function(filename) {
        serverRequest("UI.Utility","GetFileLink",filename).then(function(fileLink) {
            dom.byId("fileDownloadFrame").src = fileLink+"&prompt=true";
        });
	};
    
    // FilterView is used when the report is not interactive. It houses the filters in the main page area
    var FilterView = declare([View],{
        buildContent: function() {
            // Summary:
            //      This method is responsible for setting up the content pane's content 

            var t = this;
            t.showList.push(t.printIconNode);
        }
    });

	var ReportController = declare([Controller], {
        showAccountSummary: false,
        
        report: null,           // instance of thornberry/report/Report
        userSettings: null,     // thornberry/UserSettings
        generalUserSettings: null,  // thornberry/UserSettings
        presetManager: null,    // thornberry/PresetManager
        deviceAppId: "",

        gridView: null,         // thornberry/report/GridView
        filterWidgets: null,    // Array of thornberry/report/filter/_FilterMixin

        filterValueIndex: null, // Stores the filter values each time we call to the server. Used to compare when the next call is made so we can watch which filters have changed

		showGridOnOpen: false,
		
        search: "",

        results: null,

        updateDisableList: null, // Array of widgets to disable whenever the server is being called
        loadDeferredArray: null,
        _loadDeferred: null,
        _lastPreset: null,

        modified: null, // object

        postMixInProperties: function() {
            var t = this;
            t.inherited(arguments);

            t.set("helpText",t.generateHelp());

            t.filterWidgets = [];
            t.updateDisableList = [];

            t._loadDeferred = new Deferred();
            t.loadDeferredArray = [t._loadDeferred];

            t.filterValueIndex = {};

            t.modified = {
                any: false,
                filters: false,
                options: false,
                sort: false,
                summarization: false,
                search: false
            };
        },

        buildRendering: function() {
			var t = this;
            t.inherited(arguments);

            t._createOptionView();
            t._createFilterView();
            t._createPresetView();
        },

        postCreate: function() {
            var t = this;
            t.inherited(arguments);

            t.own(
                on(t.presetManager,"setPreset",function(preset) {
                    t.updateFromPreset(preset);
                    t.set("search","");
                    t.modified.summarization = true;
                    t.modified.filters = true;
                    t.modified.options = true;
                    t.modified.sort = true;
                    t.modified.search = true;
                    t.setModified("any");
                    domStyle.set(t.presetSaveButton.domNode,"display","none");
                    t.refresh();
                }),
                on(t.presetManager,"changePresets",function(presets) {
                    t.updatePresetDropDown(presets);
                    domStyle.set(t.presetSaveButton.domNode,"display","none");
                }),
                aspect.after(t,"pushView",function() {
                    t._lastPreset = t.presetManager.currentPreset;
                })
            );

            all(t.loadDeferredArray).then(function() {
                t.updateFromPreset(t.presetManager.currentPreset);
                t.modified.summarization = true;
                t.modified.filters = true;
                t.modified.options = true;
                t.modified.sort = true;
                t.modified.search = true;
                t.setModified("any");
                t._lastPreset = null;
            });
            
            if (t.report.hasInteractivePresentation()) {
                t.createGrid();
                t.gridView.showList.push(t.presetControlContainer);
            } else {
                t.filterView.set("title",t.report.displayName);
                t.titlePane.set("title",t.report.displayName);
                t.filterView.showList = query.NodeList();
            }

            t.updateDisableList.push(registry.byId("btnApplyFilters"));
            t.updateDisableList.push(registry.byId("btnApplyOptions"));
            t.updateDisableList.push(t.searchBox);
            if (t.report.printable()) {
                t.filterView.showList.push(t.printIconNode);
            }

            t.initializeDisplay();

            // Create tooltip object for default formatter function
            new Tooltip({
                connectId: [t.domNode],
                selector: ".tooltip-opener-node",
                getContent: function(matchedNode) {
                    return matchedNode.getAttribute("data-tooltip-text").toString().replace(/\n/g,"<br />");
                },
                showDelay: 200
            });
        },

        initializeDisplay: function() {
            var t = this;
            if (t.report.hasInteractivePresentation()) {
                t.setView(t.gridView);
                if (t.report.filterColumns !== "" && !t.showGridOnOpen) {
                    t.pushView(t.filterView);
                } else {
                    t.defer("refresh");
                }
            } else {
				if (t.report.staticColumns != 1) {
					t.filterView.showList.push(t.optionIconNode);
				}
                t.filterView.showList.push(t.summarizationIconNode);
                t.setView(t.filterView);
            }
        },

        setModified: function(itemName) {
            var t = this;
            t.modified[itemName] = true;
            t.modified.any = true;
        },

        clearModified: function() {
            var t = this;
            var itemName;
            for (itemName in t.modified) {
                t.modified[itemName] = false;
            }
        },

        createGrid: function() {
            var t = this;
            t.set("gridView",new GridView({
                controller: t,
                title: t.report.displayName
            }));
        },

        createReportManager: function() {
            var t = this;
            try {
                t.reportManager = new ReportManager({
                    report: t.report,
                    getContext: function(presentationType,report) {
                        var context = false;
                        var errorMessage = t.filterEditor.okToSave(true);
                        if (errorMessage === "") {
                            context = t.getContext(presentationType,report);
                        } else {
                            if (t.getView() !== t.filterView.id) {
                                t.pushView(t.filterView);
                            }
                            t.showMessage(errorMessage);
                        }
                        return context;
                    }
                });
                t.reportManager.startup();
            } catch(exception) {console.log(exception);}
        },
        
        updateFromPreset: function(preset) {
            var t = this;
            if (preset !== null) {
                t.presetDropDownButton.set("label",escapeText(preset.getName()));
                t.presetDropDownButton.set("title",escapeText(preset.description));

                // Apply filters
                var disableFilter = !!(preset.system && !preset.def);
                arrayUtils.forEach(t.filterEditor.fields,function(field) {
                    var filterValue = preset.filters[field.name];
                    if (typeof filterValue === "undefined" && (preset.def || field.get("disabled"))) {
                        // Reset any fields defined by the previous preset
                        var filterValue = t.filterDefaults[field.name];
                    }
                    if (typeof filterValue !== "undefined") {
                        field.deserialize(filterValue);
                    }
                    if (field.lockable) {
                        t.filterEditor.disable(field.name,typeof preset.filters[field.name] !== "undefined" && !!(preset.system && !preset.def));
                    }
                });
                t.defer(function() {
                    t.filterEditor.set("modified",false);
                });

                // Apply options
                var selectedColumns = preset.columns;
                if (selectedColumns.length === 0) {
                    // If no user default exists and the report has no default selection, then include all in view order
                    selectedColumns = arrayUtils.map(t.report.getDisplayedColumns(),function (column) {
                        return column.name;
                    });
                }
                t.columnPane.set("value",selectedColumns);
                t.defer(function() {
                    t.columnPane.set("modified",false);
                });

                if (t.gridView) {
                    // Apply sort
                    var sort = preset.sort;
                    if (sort && sort.column) {
                        t.gridView.grid.sort._sortId = sort.column;
                        t.gridView.grid.sort._sortDescend = (sort.direction === "DESC");
                        t.gridView.grid.header.refresh();
                    }
                }
            }
        },
        
        getContext: function(presentationType,/*thornberry/report/Report*/ report) {
            // Summary:
            //      Gathers all context information for the current report including
            //          - filters
            //          - columns
            //          - sorting
            //          - summarization
            //          - custom report options
            var t = this;
            var context = {
                userId: t.userSettings.userId,
                deviceAppId: t.deviceAppId,
                presentation: presentationType,
                reportClass: report.className,
                filters: t.getFilterContext(),
                columns: t.columnPane.get("value"),
                summarizeBy: "",
                search: t.get("search"),
                options: {},
				audit: t.audit(), // boolean; should be true whenever the filters or columns change, but not for sorting or requests for new pages
				filterFlag: t.filterFlag() // boolean; should be true only when the filters change
            };
            if (t.gridView) {
                if (t.gridView.getSort) {
                    context.sort = t.gridView.getSort();
                }

                if (t.report.selectableRows) {
                    var column = t.report.getColumn(t.report.primaryKeyColumn);
                    var selectedIds = t.gridView.getSelected();
                    if (!column && selectedIds.length === 0 && report.childReport) {
                        t.showMessage("Please make a selection from the main interactive view first");
                        return false;
                    }
                    if (column) {
                        var displayValue = arrayUtils.map(selectedIds,function(id) {
                            return t.gridView.dataStore.get(id)[t.report.summaryColumn];
                        }).join(", ");

                        context.filters.push({
                            column: "selectedIds",
                            displayName: "Selected "+column.displayName+"(s)",
                            value: selectedIds,
                            displayValue: displayValue
                        });
                    } else {
                        context.options.selectedIds = selectedIds;
                    }
                }
            }
            if (!t.presetManager.currentPreset.def && report.className === t.report.className) {
                context.options.presetName = t.presetManager.currentPreset.name;
            }
            return context;
        },

        getFilterContext: function() {
            // Summary:
            //      Cycles through all filter widgets and collects values to be passed to the server
            var t = this;
            var filters = [];
            arrayUtils.forEach(t.filterWidgets,function(filterWidget) {
                var oldValue = t.filterValueIndex[filterWidget.name];
                var newValue = filterWidget.serialize();
                var changed = (typeof oldValue === "undefined" || oldValue != newValue);

                filters.push({
                    column: filterWidget.columnDefinition.name,
                    displayName: (filterWidget.columnDefinition.filterName ? filterWidget.columnDefinition.filterName : filterWidget.columnDefinition.displayName),
                    value: filterWidget.getFilterValue(),
                    displayValue: filterWidget.getDisplayValue(),
                    changed: changed // whether the filter has changed since the last server call
                });

                t.filterValueIndex[filterWidget.name] = newValue;
            });
            return filters;
        },

        audit: function() {
            var t = this;
            return t.modified.filters || t.modified.columns || t.modified.search;
        },
		
		filterFlag: function() {
            var t = this;
            return t.modified.filters;
        },

        _setGridViewAttr: function(gridView) {
            var t = this;
            t.gridView = gridView;
            t.addView(gridView);
        },

        _createOptionView: function() {
            // Summary:
            //      Creates a thornberry/layout/DragAndDropPane for users to control which columns get displayed in the interactive view
            var t = this;
            t.columnPane = new DragAndDropPane({
                style: "width:100%;height:100%",
                store: new Memory({data:t.report.getDisplayedColumns(),idProperty:"name"}),
                sort: [{attribute:"name"}],
                value: [],
                idAttr: "name",
                labelAttr: "displayName"
            });

            domConstruct.place(t.columnPane.domNode,t.optionContainer);
            t.columnPane.startup();

            t.optionView.isModified = function() {
                return t.columnPane.get("modified");
            };
        },

        _createFilterView: function() {
            var t = this;

            t.filterView.unlockButton = new Button({
                label: "Unlock Filters",
                onClick: function() {
                    t.presetManager.copyToMyDefault();
                }
            }).placeAt(t.actionContainer);
            t.registerToggledNode(t.filterView.unlockButton.domNode);
            t.filterView.refresh = function() {
                var t = this;
                if (!t.controller.presetManager.currentPreset.def) {
                    //domStyle.set(t.unlockButton.domNode,"display","");
                }
            };

            t.filterEditor = new ObjectEditor({
                templateString: template.autoGenerateFilterTemplate(t.report),
                report: t.report,
                controller: t,
                object: new Stateful({}),
                objectModel: new ObjectModel(),

                buildRendering: function() {
                    var t = this;
                    t.inherited("buildRendering",arguments);

                    if (t.report) {
                        var fieldDeferredArray = [];
                        arrayUtils.forEach(t.report.filterColumns.split(","),function(column) {
                            var fieldDeferred = new Deferred();
                            fieldDeferredArray.push(fieldDeferred);

                            try {
                                var columnDefinition = t.report.getColumn(column);
                                if (columnDefinition && columnDefinition.isFilterable && columnDefinition.filterWidget) {
                                    require([columnDefinition.filterWidget],function(FilterWidget) {
                                        var nodeName = columnDefinition.name+"Node";
                                        var filterWidget = new FilterWidget({columnDefinition:columnDefinition},t[nodeName]);
                                        if (columnDefinition.hasOptions) {
                                            filterWidget.set("optionsValue",t.controller.filterOptions[columnDefinition.name]);
                                        }
                                        filterWidget.startup();
                                        t.controller.filterWidgets.push(filterWidget);

                                        var fieldName = columnDefinition.name+"Field";
                                        t[fieldName] = filterWidget;
                                        filterWidget.name = columnDefinition.name;
                                        t.own(on(t[fieldName],"change",function(value) {t._modifyValue(this.name);}));
                                        t.fields.push(t[fieldName]);

                                        if (columnDefinition.filterRequired) {
                                            t.require(columnDefinition.name);
                                        }
                                        fieldDeferred.resolve();
                                    });
                                } else {
                                    fieldDeferred.resolve();
                                }
                            } catch (exception) {
                                console.log(exception);
                                fieldDeferred.resolve();
                            }
                        });
                        t.controller.loadDeferredArray.push(all(fieldDeferredArray));
                    }
                },

                postCreate: function() {
                    var t = this;
                    t.inherited("postCreate",arguments);

                    var requiredFilters = t.controller.requiredFilters;
                    /*if (requiredFilters !== "null") {
                        arrayUtils.forEach(t.controller.plugin.requiredFilters,function(filter) {
                            t.require(filter);
                        }); 
                    }*/
                },

                okToSave: function(suppressMessage) {
                    var t = this;
                    var errorMessage = "";
                    
                    if (!t.formNode.validate()) {
                        errorMessage = "Please correct any invalid fields.";
                    } else if (query(".pinkBackground",t.domNode).length > 0) {
                        errorMessage = "Please fill in all required fields.";
                    }
                    if (errorMessage !== "" && !suppressMessage) {
                        actionShowMessage(errorMessage);
                    }
                    return errorMessage;
                }
            });

            t.filterContainer.appendChild(t.filterEditor.domNode);
            t.filterEditor._postModify();

            t.filterView.isModified = function() {
                return t.filterEditor.get("modified") || !t.filterEditor.formNode.validate();
            };
        },

        _createPresetView: function() {
            var t = this;
            var presetControlContainer = domConstruct.toDom("<div style='position:relative;margin:auto;width:350px'></div>");
            t.actionPane.domNode.appendChild(presetControlContainer);
            //t.registerToggledNode(presetControlContainer);
            t.presetControlContainer = presetControlContainer;

            t.presetMenu = new DropDownMenu({style:"display:none"});
            
            t.presetMenu.startup();
            t.presetDropDownButton = new DropDownButton({
                label: t.presetManager.currentPreset.name,
                title: t.presetManager.currentPreset.description,
                dropDown: t.presetMenu
            }).placeAt(presetControlContainer);
            t.presetDropDownButton.startup();
            
            t.updatePresetDropDown();

            /*t.presetViewButton = new Button({
                label: "View",
                title: "View presets",
                onClick: function() {

                }
            }).placeAt(presetControlContainer);*/

            t.presetSaveButton = new Button({
                label: "Save",
                title: "Save as My Default",
                onClick: function() {
                    t.showMessage("Save the current configuration as My Default?",[
                        {label:"Yes",callback:lang.hitch(t.presetManager,"saveMyDefault")},
                        {label:"No"}
                    ]);
                },
                style: "display:none"
            }).placeAt(presetControlContainer);
        },

        updatePresetDropDown: function() {
            var t = this;
            var presets = t.presetManager.getPresetsSorted();
            t.presetMenu.destroyDescendants();
            arrayUtils.forEach(presets,function(preset) {
                var menuItem = new MenuItem({
                    label: escapeText(preset.getName()),
                    title: escapeText(preset.description),
                    onClick: function() {
                        t.presetManager.set("currentPreset",preset);
                    }
                });
                t.presetMenu.addChild(menuItem);
            });
            t.presetDropDownButton.set("label",t.presetManager.currentPreset.getName());
        },

        _updateServerCalls: function() {
            var t = this;
            t.inherited(arguments);

            try {
                // Disabled widgets in the list when a server call is in progress
                var disabled = (t._serverCalls > 0);
                arrayUtils.forEach(t.updateDisableList,function(widget) {
                    if (widget) {
                        widget.set("disabled",disabled);
                    }
                });
            } catch (exception) {}
        },

        _setSearchAttr: function(value) {
            var t = this;
            if (value !== t.search) {
                t.search = value;
                t.setModified("search");
            }
        },
        
        onApplyOptions: function() {
            var t = this;
            t.set("search","");
            if (t.columnPane.get("modified")) {
                var columns = t.columnPane.get("value");

                var currentPreset = t.presetManager.currentPreset;
                currentPreset.columns = columns;
                t.setCurrentPresetModified();

                t.columnPane.set("modified",false);
                t.setModified("options");
            }
            if (t.getView() === t.optionView.id) {
                t.popView();
            }
        },
        onCancelOptions: function() {
            var t = this;
            t.onCancel(t.optionView);
            t.columnPane.set("modified",false);
            t.popView();
        },

        onApplyFilters: function() {
            var t = this;
            if (!t.filterEditor || t.filterEditor.okToSave() === "") {
                t.set("search","");

                if (t.filterEditor.get("modified")) {
                    // Copy to current preset
                    var currentPreset = t.presetManager.currentPreset;
                    arrayUtils.forEach(t.filterEditor.fields,function(field) {
                        currentPreset.filters[field.name] = field.serialize();
                    });
                    t.setCurrentPresetModified();
                    t.filterEditor.set("modified",false);
                    t.setModified("filters");
                }

                if (t.getView() === t.filterView.id) {
                    t.popView();
                }
            }
        },
        onCancelFilters: function() {
            var t = this;
            t.onCancel(t.filterView);
            t.filterEditor.set("modified",false);
            t.popView();
        },
        onSearch: function(value) {
            var t = this;
            if (value !== "") {
                t.set("search",value);
            }
        },
        onCancel: function(view) {
            var t = this;
            if (t.modified.any && t._lastPreset) {
                t.presetManager.set("currentPreset",t._lastPreset);
                t.clearModified();
            }
            if (view.isModified()) {
                t.updateFromPreset(t.presetManager.currentPreset);
            }
        },
        
        setCurrentPresetModified: function() {
            var t  = this;
            var currentPreset = t.presetManager.currentPreset;
            currentPreset.set("modified",true);
            t.presetDropDownButton.set("label",currentPreset.getName());
            domStyle.set(t.presetSaveButton.domNode,"display","");
        },

        startup: function(value) {
            var t = this;
            t.inherited(arguments);
            t._loadDeferred.resolve();
        },

        generateHelp: function () {
            var report = this.report;

            var replacement = {};

            replacement.reportTitle = escapeText(report.displayName);
            replacement.reportDescription = report.description; // Don't escape this value; HTML may be used for formatting the long description
            replacement.reportHelp = report.help;

            replacement.columnHelp = "<ul>";
            arrayUtils.forEach(report.columnDefinitions, function (columnDefinition) {
                if (columnDefinition.isDisplayed) {
                    replacement.columnHelp += "<li><span class='bold'>" + columnDefinition.displayName + "</span>" + (columnDefinition.help !== "" ? " - " : "") + escapeText(columnDefinition.help) + "</li>";
                }
            });
            replacement.columnHelp += "</ul>";

            replacement.filterHelp = "<ul>";
            arrayUtils.forEach(report.filterColumns.toString().split(","), function (column) {
                columnDefinition = report.getColumn(column);
                if (columnDefinition && columnDefinition.filterHelp != null) {
                    replacement.filterHelp += "<li><span class='bold'>" + (columnDefinition.filterName ? columnDefinition.filterName : columnDefinition.displayName) + "</span>" + (columnDefinition.filterHelp !== "" ? " - " : "") + escapeText(columnDefinition.filterHelp) + "</li>";
                }
            });
            replacement.filterHelp += "</ul>";

            replacement.sqlQuery = escapeText(report.queryText);
            replacement.viewName = escapeText(report.viewName);

            replacement.presetHelp = "";
            // Configurations
            var presetManager = this.presetManager;
            if (presetManager && presetManager.presetData.length > 0) {
                replacement.presetHelp += "<table class='tbl-data-std tbl-alt'>";
                replacement.presetHelp +=   "<thead>";
                replacement.presetHelp +=     "<tr>";
                replacement.presetHelp +=       "<th>Name</th>";
                replacement.presetHelp +=       "<th>Description</th>";
                replacement.presetHelp +=     "</tr>";
                replacement.presetHelp +=   "</thead>";
                replacement.presetHelp +=   "<tbody>";
                var presets = presetManager.getPresetsSorted();
                arrayUtils.forEach(presets, function (preset) {
                    replacement.presetHelp += "<tr>";
                    replacement.presetHelp +=   "<td>" + escapeText(preset.name) + "</td>";
                    replacement.presetHelp +=   "<td>" + escapeText(preset.description) + "</td>";
                    replacement.presetHelp += "</tr>";
                });
                replacement.presetHelp +=   "</tbody>";
                replacement.presetHelp += "</table>";
            }

            if (replacement.presetHelp !== "") {
                replacement.presetHelp = "<h4 id='presetConfigHelp'>Configurations</h4>" + replacement.presetHelp;
            }

            return lang.replace(helpTemplate, replacement);
        }, 
	});
	
	return ReportController;
});