define(["dojo/Stateful","dojo/_base/declare","dojo/_base/lang","dojo/date","dojo/date/locale","thornberry/volunteer/utils"],function(Stateful,declare,lang,date,locale,utils) {
	var Activity = declare(Stateful, {
		
		// Properties from database
		id: "",
		guid: "",
		visitGuid: "",
		volunteerId: "",
		teamId: "",
		typeId: "",
		startDate: "",
		startTime: "",
		activityTime: "",
		travelTime: "",
		mileage: "",
		patientId: "",
		companyId: "",
		locationId: "",
		active: 1,
		comment: "",
		
		// Additional properties
		endDate: "",
		endTime: "",
		totalTime: 0,
		
		// Object references
		volunteer: null,
		team: null,
		type: null,
		company: null,
		location: null,
		patient: null,
		
		// Display values
		volunteerName: "",
		patientName: "",
		patientLink: "",
		activityName: "",
		startDateTime: "",
		activityTimeDisplay: "",
		travelTimeDisplay: "",
		totalTimeDisplay: "",
		
		watcher: null,
		watchHandle: null,
		checkForOverlap: false,
		
		constructor: function(/*Object*/ activityData){
			var t = this;
			lang.mixin(t,activityData);
			//t.inherited(activityData);
			if (!t.id) {
				t.id = Activity.counter++; // Assign a tempId
			}
			
			t.watch("startDate",function(name,oldValue,value) {
				t._updateStartDateTime();
			});
			t.watch("startTime",function(name,oldValue,value) {
				t._updateStartDateTime();
			});
			t.watch("activityTime",function(name,oldValue,value) {
				t.set("activityTimeDisplay",utils.formatMinutes(value));
			});
			t.watch("travelTime",function(name,oldValue,value) {
				t.set("travelTimeDisplay",utils.formatMinutes(value));
			});
			t.watch("totalTime",function(name,oldValue,value) {
				t.set("totalTimeDisplay",utils.formatMinutes(value));
			});
		},
		
		getSaveData: function(action) {
			var t = this;
			return {
				id: ((action !== "Add" && action !== "Copy") ? t.id : Activity.counter++),
				volunteerId: t.volunteerId,
				teamId: t.teamId,
				typeId: t.typeId,
				startDate: t.startDate,
				startTime: t.startTime,
				activityTime: t.activityTime,
				travelTime: t.travelTime,
				mileage: t.mileage,
				patientId: t.patientId,
				companyId: t.companyId,
				locationId: t.locationId,
				active: t.active,
				comment: t.comment
			};
		},
		
		_watcherSetter: function(value) {
			var t = this;
			if (t.watchHandle) {
				t.watchHandle.unwatch();
				t.watchHandle = null;
			}
			
			t.watcher = value;
			if (value) {
				t.watchHandle = t.watch(function(name,oldValue,newValue) {
					t.watcher.notify(name,newValue);
				});
			}
		},
		
		startDateGetDisplay: function() {
			return utils.parseDate(this.startDate);
		},
		
		startTimeGetDisplay: function() {
			return utils.parseTime(this.startTime);
		},
		
		endDateGetDisplay: function() {
			return utils.parseDate(this.endDate);
		},
		
		endTimeGetDisplay: function() {
			return utils.parseTime(this.endTime);
		},
		
		activityTimeGetDisplay: function() {
			return this.activityTimeDisplay;
		},
		
		travelTimeGetDisplay: function() {
			return this.travelTimeDisplay;
		},
		
		totalTimeGetDisplay: function() {
			return this.totalTimeDisplay;
		},
		
		getStart: function() {
			var t = this;
			return locale.parse(t.startDate,{selector:"date",datePattern:"yyyy-MM-dd"});
		},
		
		_startDateSetter: function(value) {
			var t = this;
			t.startDate = value;
			if (value) {
				t._updateTimes("start");
			}
		},
		
		_startTimeSetter: function(value) {
			var t = this;
			t.startTime = value;
			if (value) {
				t._updateTimes("start");
			}
		},
		
		_endDateSetter: function(value) {
			var t = this;
			t.endDate = value;
			if (value) {
				t._updateTimes("end");
			}
		},
		
		_endTimeSetter: function(value) {
			var t = this;
			t.endTime = value;
			if (value) {
				t._updateTimes("end");
			}
		},
		
		_activityTimeSetter: function(value) {
			var t = this;
			if (value !== "") {
				value = parseInt(value);
			}
			t.activityTime = value;
			t._updateTimes("time");
		},
		
		_travelTimeSetter: function(value) {
			var t = this;
			if (value !== "") {
				value = parseInt(value);
			}
			t.travelTime = value;
			t._updateTimes("travel");
		},
		
		_totalTimeSetter: function(value) {
			var t = this;
			t.totalTime = value;
		},
		
		_updateTimes: function(changeFrom) {
			// date - yyyy-MM-dd
			// time - HHmm
			// time 2 - integer minutes
			var t = this;
			
			t.set("checkForOverlap",false);
			var checkForOverlap = false;
			
			var start = utils.parseDateAndTime(t.startDate,t.startTime);
			var end = utils.parseDateAndTime(t.endDate,t.endTime);
			var activityTime = t.activityTime;
			
			// enough data to calculate all dates?
			var dateChange = false;
			var startDate = "";
			var startTime = "";
			var endDate = "";
			var endTime = "";
			if (!(changeFrom === "travel") && ((start && end) || (start && activityTime !== "") || (end && activityTime !== ""))) {
				dateChange = true;
				
				if (changeFrom === "start") {
					if (activityTime !== "") {
						end = date.add(start,"minute",activityTime);
					} else {
						activityTime = date.difference(start,end,"minute");
					}
				} else if (changeFrom === "end") {
					if (start) {
						activityTime = date.difference(start,end,"minute");
					} else {
						start = date.add(end,"minute",-activityTime);
					}
				} else if (changeFrom === "time") {
					if (activityTime === "") {
						activityTime = 0;
					}
					if (start) {
						end = date.add(start,"minute",activityTime);
					} else {
						start = date.add(end,"minute",-activityTime);
					}
				}
				
				startDate = locale.format(start,{datePattern:"yyyy-MM-dd", selector:"date"});
				startTime = locale.format(start,{selector:"time",timePattern:"HHmm"});
				endDate = locale.format(end,{datePattern:"yyyy-MM-dd", selector:"date"});
				endTime = locale.format(end,{selector:"time",timePattern:"HHmm"});
			}
			
			var totalTime = (activityTime !== "" ? parseInt(activityTime) : 0) + (t.travelTime !== "" ? parseInt(t.travelTime) : 0);
			
			if (dateChange) {
				t._changeAttrValue("startDate",startDate);
				t._changeAttrValue("startTime",startTime);
				t._changeAttrValue("endDate",endDate);
				t._changeAttrValue("endTime",endTime);
				t._changeAttrValue("activityTime",activityTime);
				t._updateStartDateTime();
			}
			t.set("totalTime",totalTime);
			
			t.set("checkForOverlap",true);
		},
		
		_updateStartDateTime: function() {
			var t = this;
			var val = "";
			if (t.startTime && t.startDate) {
				var date = locale.parse(t.startDate+" "+t.startTime,{selector:"date",datePattern:"yyyy-MM-dd HHmm"});
				val = locale.format(date,{datePattern:"MM/dd/yyyy HHmm", selector:"date"});
			} else if (t.startDate) {
				var date = locale.parse(t.startDate,{selector:"date",datePattern:"yyyy-MM-dd"});
				val = locale.format(date,{datePattern:"MM/dd/yyyy", selector:"date"});
			}
			t.set("startDateTime",val);
		},
		
		_patientSetter: function(value) {
			var t = this;
			t.patient = value;
			if (value) {
				t._changeAttrValue("patientId",value.id);
				t.set("patientName",value.name+" ("+value.pid+")");
				t.set("patientLink","<a onclick=\"startWindow.OpenSummary('"+value.id+"');window.focusOpen('','STARTWINDOW').focus()\">"+value.name+" ("+value.pid+")</a>");
				utils.stores.patientStore.put(value); // Keep the patient store up to date with additional patients from the server sent within activities
			} else {
				t._changeAttrValue("patientId","");
				t.set("patientName","");
				t.set("patientLink","");
			}
		},
		
		_patientIdSetter: function(value) {
			var t = this;
			t.patientId = value;
			var patient = utils.stores.patientStore.get(value)
			if (patient) {
				t.set("patient",patient);
			} else {
				t.set("patient",null);
			}
		},
		
		_volunteerIdSetter: function(value) {
			var t = this;
			t.volunteerId = value;
			var volunteer = utils.stores.volunteerStore.get(value);
			if (volunteer) {
				t.set("volunteer",volunteer);
				t.set("volunteerName",volunteer.name);
			} else {
				t.set("volunteer",null);
				t.set("volunteerName","");
			}
		},
		
		_teamIdSetter: function(value) {
			var t = this;
			t.teamId = value;
			t.set("team",utils.stores.teamStore.get(value));
		},
		
		_typeIdSetter: function(value) {
			var t = this;
			t.typeId = value;
			var type = utils.stores.typeStore.get(value);
			if (type) {
				t.set("type",type);
				t.set("activityName",type.name);
			} else {
				t.set("type","");
				t.set("activityName","");
			}
		},
		
		_companyIdSetter: function(value) {
			var t = this;
			t.companyId = value;
			t.set("company",utils.stores.companyStore.get(value));
		},
		
		_locationIdSetter: function(value) {
			var t = this;
			t.locationId = value;
			t.set("location",utils.stores.locationStore.get(value));
		}
	});
	
	Activity.counter = 0;
	
	return Activity;
});